﻿using P5T.Models;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Mvc;
using System.IO.Compression;
using System.IO;
using Ionic.Zip;

namespace P5T.Controllers
{
    [AuthFilter]
    public class ProgramsController : Controller
    {
        [HttpGet]
        public ActionResult View(int id)
        {
            var currentUser = this.RouteData.DataTokens["user"] as ResearcherModel;

            var program = ProgramModel.Find(id);
            if (program == null)
            {
                return RedirectToAction("List", "Programs");
            }

            ViewBag.User = currentUser;

            return View(program);
        }

        [HttpGet]
        public ActionResult List()
        {
            return View(ProgramModel.FindAll(false));
        }

        [HttpGet]
        public ActionResult Create()
        {
            return View();
        }

        [HttpPost]
        public ActionResult Create(string name, string description)
        {
            // Get the current user
            var currentUser = this.RouteData.DataTokens["user"] as ResearcherModel;

            // Create the actual program.
            var program = ProgramModel.Create(name, description);
            
            // Log that it was created.
            ProgramActivityModel.CreateCreateActivity(DateTime.Now, currentUser.Id, program.Id);

            // Subscribe the creator to it
            ProgramSubscriptionModel.Create(program.Id, currentUser.Id);

            return RedirectToAction("View", "Programs", new { id = program.Id });
        }

        [HttpGet]
        public ActionResult Edit(int id, bool? ajax)
        {
            ViewBag.ajax = ajax.HasValue && ajax.Value;

            var program = ProgramModel.Find(id);
            if (program == null)
            {
                return RedirectToAction("List", "Programs");
            }
            return View(program);
        }

        [HttpPost]
        public ActionResult Edit(int id, string name, string description)
        {
            // Find the current user
            var currentUser = this.RouteData.DataTokens["user"] as ResearcherModel;

            // Find the appropriate programModel.
            var program = ProgramModel.Find(id);
            if (program != null)
            {
                // Make appropriate changes, recording old dat as needed.
                string old_name = null;

                bool updatedName = false;
                bool updatedDesc = false;
                if (program.Name != name)
                {
                    old_name = program.Name;
                    program.Name = name;
                    updatedName = true;
                }
                if (program.Description != description)
                {
                    program.Description = description;
                    updatedDesc = true;
                }

                // Perform the update
                if (updatedName || updatedDesc)
                {
                    ProgramModel.Update(program);

                    // Log the change to the name, if needed.
                    if (updatedName)
                    {
                        ProgramActivityModel.CreateRenamedActivity(DateTime.Now, currentUser.Id, program.Id, old_name, name);
                    }

                    // Log the change to the description, if needed.
                    if (updatedDesc)
                    {
                        ProgramActivityModel.CreateChangedDescActivity(DateTime.Now, currentUser.Id, program.Id, description);
                    }
                }

                return Json(new {success = true});
            }
            return HttpNotFound();
        }

        [HttpGet]
        public ActionResult Delete(int id)
        {
            var program = ProgramModel.Find(id);
            if (program != null)
            {
                return View(program);
            }
            return RedirectToAction("List", "Programs");
        }

        [HttpPost]
        public ActionResult DoDelete(int id)
        {
            //Get the current user
            var currentUser = this.RouteData.DataTokens["user"] as ResearcherModel;

            ProgramModel.Delete(id);

            ProgramActivityModel.CreateArchivedActivity(DateTime.Now, currentUser.Id, id);

            return Json(new { success = true });
        }

        [HttpPost]
        public ActionResult FetchTitleAndDesc(int id)
        {
            var program = ProgramModel.Find(id);
            if(program != null)
            {
                return Json(new { title = program.Name, description = program.Description });
            }
            return HttpNotFound();
        }

        [HttpGet]
        public ActionResult FetchResearchArticles(int id)
        {
            var program = ProgramModel.Find(id);
            if (program != null)
            {
                return PartialView("~/Views/Shared/Artifacts/View/ResearchArticle.cshtml", program);
            }
            return HttpNotFound();
        }

        [HttpGet]
        public ActionResult FetchJournalArticles(int id)
        {
            var program = ProgramModel.Find(id);
            if (program != null)
            {
                return PartialView("~/Views/Shared/Artifacts/View/JournalArticle.cshtml", program);
            }
            return HttpNotFound();
        }

        [HttpGet]
        public ActionResult FetchConferencePresentations(int id)
        {
            var program = ProgramModel.Find(id);
            if (program != null)
            {
                return PartialView("~/Views/Shared/Artifacts/View/ConferencePresentation.cshtml", program);
            }
            return HttpNotFound();
        }

        [HttpGet]
        public ActionResult FetchMediaInterviews(int id)
        {
            var program = ProgramModel.Find(id);
            if (program != null)
            {
                return PartialView("~/Views/Shared/Artifacts/View/MediaInterview.cshtml", program);
            }
            return HttpNotFound();
        }

        [HttpGet]
        public ActionResult FetchGrants(int id)
        {
            var program = ProgramModel.Find(id);
            if (program != null)
            {
                return PartialView("~/Views/Shared/Artifacts/View/Grant.cshtml", program);
            }
            return HttpNotFound();
        }

        [HttpGet]
        public ActionResult FetchOthers(int id)
        {
            var program = ProgramModel.Find(id);
            if (program != null)
            {
                return PartialView("~/Views/Shared/Artifacts/View/Other.cshtml", program);
            }
            return HttpNotFound();
        }

        [HttpGet]
        public ActionResult Watch(int program_id)
        {
            var program = ProgramModel.Find(program_id);
            if (program != null)
            {
                var currentUser = this.RouteData.DataTokens["user"] as ResearcherModel;
                ProgramSubscriptionModel.Create(program_id, currentUser.Id);
                ViewBag.User = currentUser;
                return View("~/Views/Programs/WatchButton.cshtml", program);
            }
            return HttpNotFound();
        }

        [HttpGet]
        public ActionResult Unwatch(int program_id)
        {
            var program = ProgramModel.Find(program_id);
            if (program != null)
            {
                var currentUser = this.RouteData.DataTokens["user"] as ResearcherModel;
                ProgramSubscriptionModel.Delete(program_id, currentUser.Id);
                ViewBag.User = currentUser;
                return View("~/Views/Programs/WatchButton.cshtml", program);
            }
            return HttpNotFound();
        }

        [HttpGet]
        public ActionResult History(int program_id)
        {
            ViewBag.specific = true;
            return View("~/Views/Programs/ViewHistory.cshtml", ProgramActivityModel.FindAllByArtifactId(program_id));
        }

        [HttpGet]
        public ActionResult DownloadArtifacts(int id)
        {
            var program = ProgramModel.Find(id);

            if(program != null)
            {
                var outputStream = new MemoryStream();
                using (var zip = new ZipFile())
                {
                    foreach (var ra in program.ResearchArticles)
                    {
                        // Write out metadata.
                        zip.AddEntry("research_articles/" + ra.Id.ToString() + "/metadata.txt", ra.GetMetadata());

                        // Add Files
                        foreach (var file in ra.Attachments)
                        {
                            zip.AddEntry("research_articles/" + ra.Id.ToString() + "/" + file.Id.ToString() + " - " + file.Name, file.Data.Value);
                        }

                        // Add Urls
                        String urls = "";
                        foreach (var url in ra.Urls)
                        {
                            urls += String.Format("{0}\n", url.Url);
                        }
                        zip.AddEntry("research_articles/" + ra.Id.ToString() + "/urls.txt", urls);
                    }

                    foreach (var ja in program.JournalArticles)
                    {
                        // Write out metadata.
                        zip.AddEntry("journal_articles/" + ja.Id.ToString() + "/metadata.txt", ja.GetMetadata());

                        // Add Files
                        foreach (var file in ja.Attachments)
                        {
                            zip.AddEntry("journal_articles/" + ja.Id.ToString() + "/" + file.Id.ToString() + " - " + file.Name, file.Data.Value);
                        }

                        // Add Urls
                        String urls = "";
                        foreach (var url in ja.Urls)
                        {
                            urls += String.Format("{0}\n", url.Url);
                        }
                        zip.AddEntry("journal_articles/" + ja.Id.ToString() + "/urls.txt", urls);
                    }

                    foreach (var mi in program.MediaInterviews)
                    {
                        // Write out metadata.
                        zip.AddEntry("media_interviews/" + mi.Id.ToString() + "/metadata.txt", mi.GetMetadata());

                        // Add Files
                        foreach(var file in mi.Attachments)
                        {
                            zip.AddEntry("media_interviews/" + mi.Id.ToString() + "/" + file.Id.ToString() + " - " + file.Name, file.Data.Value);
                        }

                        // Add Urls
                        String urls = "";
                        foreach (var url in mi.Urls)
                        {
                            urls += String.Format("{0}\n", url.Url);
                        }
                        zip.AddEntry("media_interviews/" + mi.Id.ToString() + "/urls.txt", urls);
                    }

                    foreach (var cp in program.ConferencePresentations)
                    {
                        // Write out metadata.
                        zip.AddEntry("conference_presentations/" + cp.Id.ToString() + "/metadata.txt", cp.GetMetadata());

                        // Add Files
                        foreach (var file in cp.Attachments)
                        {
                            zip.AddEntry("conference_presentations/" + cp.Id.ToString() + "/" + file.Id.ToString() + " - " + file.Name, file.Data.Value);
                        }

                        // Add Urls
                        String urls = "";
                        foreach (var url in cp.Urls)
                        {
                            urls += String.Format("{0}\n", url.Url);
                        }
                        zip.AddEntry("conference_presentations/" + cp.Id.ToString() + "/urls.txt", urls);
                    }

                    foreach (var gr in program.Grants)
                    {
                        // Write out metadata.
                        zip.AddEntry("grants/" + gr.Id.ToString() + "/metadata.txt", gr.GetMetadata());

                        // Add Files
                        foreach (var file in gr.Attachments)
                        {
                            zip.AddEntry("grants/" + gr.Id.ToString() + "/" + file.Id.ToString() + " - " + file.Name, file.Data.Value);
                        }

                        // Add Urls
                        String urls = "";
                        foreach (var url in gr.Urls)
                        {
                            urls += String.Format("{0}\n", url.Url);
                        }
                        zip.AddEntry("grants/" + gr.Id.ToString() + "/urls.txt", urls);
                    }

                    foreach (var ot in program.Others)
                    {
                        // Write out metadata.
                        zip.AddEntry("others/" + ot.Id.ToString() + "/metadata.txt", ot.GetMetadata());

                        // Add Files
                        foreach (var file in ot.Attachments)
                        {
                            zip.AddEntry("others/" + ot.Id.ToString() + "/" + file.Id.ToString() + " - " + file.Name, file.Data.Value);
                        }

                        // Add Urls
                        String urls = "";
                        foreach (var url in ot.Urls)
                        {
                            urls += String.Format("{0}\n", url.Url);
                        }
                        zip.AddEntry("others/" + ot.Id.ToString() + "/urls.txt", urls);
                    }

                    zip.Save(outputStream);
                }
                outputStream.Position = 0;
                return File(outputStream, System.Net.Mime.MediaTypeNames.Application.Octet, program.Name + ".zip");
            }
            return HttpNotFound();
        }
    }
}
