﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using System.Linq;
using System.Web;

namespace P5T.Models
{
    public abstract class ActivityModel
    {
        public enum ActivityModelKind
        {
            Program,
            ResearchArticle,
            JournalArticle,
            ConferencePresentation,
            MediaInterview,
            Grant,
            Other
        }

        public int Id;
        public DateTime Dt;
        public Lazy<ResearcherModel> User;
        public int Type;
        public String Param1Data;
        public String Param2Data;

        public ActivityModel(int id, DateTime dt, int researcher, int type, String param1_data, String param2_data)
        {
            this.Id = id;
            this.Dt = dt;
            this.User = new Lazy<ResearcherModel>(delegate() { return ResearcherModel.Find(researcher); }, researcher);
            this.Type = type;
            this.Param1Data = param1_data;
            this.Param2Data = param2_data;
        }

        public String GetUserName()
        {
            return User.Value.Author.Value.ToString();
        }

        public abstract ActivityModelKind GetKind();
    }

    public class ProgramActivityModel : ActivityModel
    {
        public const int CREATE = 1;
        public const int ARCHIVED = 2;
        public const int RENAMED = 3;
        public const int CHANGED_DESC = 4;

        public Lazy<ProgramModel> Program;

        public ProgramActivityModel(int id, DateTime dt, int researcher, int program, int type, String param1_data, String param2_data)
            : base(id, dt, researcher, type, param1_data, param2_data)
        {
            this.Program = new Lazy<ProgramModel>(delegate() { return ProgramModel.Find(program); }, program);
        }

        public static ProgramActivityModel Create(DateTime dt, int researcher, int program, int type, String param1_data, String param2_data)
        {
            var result = DatabaseFacade.ExecuteTypedTx<ProgramActivityModel>(delegate (SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                // First create the activity.
                cmd.CommandText = "INSERT INTO ProgramActivity (date_time, researcher_id, program_id, activity_type, param1_data, param2_data) OUTPUT INSERTED.log_id VALUES (@DateTime, @ResearcherId, @ProgramId, @ActivityType, @Param1Data, @Param2Data)";
                cmd.Parameters.Add("@DateTime", SqlDbType.DateTime).Value = dt;
                cmd.Parameters.Add("@ResearcherId", SqlDbType.Int).Value = researcher;
                cmd.Parameters.Add("@ProgramId", SqlDbType.Int).Value = program;
                cmd.Parameters.Add("@ActivityType", SqlDbType.Int).Value = type;
                cmd.Parameters.Add("@Param1Data", SqlDbType.NVarChar, 128).Value = (param1_data == null) ? SqlString.Null : param1_data;
                cmd.Parameters.Add("@Param2Data", SqlDbType.NVarChar, 128).Value = (param2_data == null) ? SqlString.Null : param2_data;
                cmd.Prepare();
                return new ProgramActivityModel((Int32)cmd.ExecuteScalar(), dt, researcher, program, type, param1_data, param2_data);
            });

            var subs = new List<SubscriptionModel>();
            subs.AddRange(UserSubscriptionModel.FindAllSubscribers(researcher));
            subs.AddRange(ProgramSubscriptionModel.FindAllSubscribers(program));

            foreach (var sub in subs)
            {
                int value = (int)sub.Subscriber.RawValue;
                if (value != researcher)
                {
                    ProgramNotificationModel.Create(value, result.Id);
                }
            }

            return result;
        }

        public static ProgramActivityModel CreateCreateActivity(DateTime dt, int researcher, int program) 
        {
            return Create(dt, researcher, program, CREATE, null, null);
        }

        public static ProgramActivityModel CreateArchivedActivity(DateTime dt, int researcher, int program) 
        {
            return Create(dt, researcher, program, ARCHIVED, null, null);
        }

        public static ProgramActivityModel CreateRenamedActivity(DateTime dt, int researcher, int program, String old_name, String new_name) 
        {
            return Create(dt, researcher, program, RENAMED, new_name, old_name);
        }

        public static ProgramActivityModel CreateChangedDescActivity(DateTime dt, int researcher, int program, String new_desc) 
        {
            return Create(dt, researcher, program, CHANGED_DESC, new_desc, null);
        }

        public static ProgramActivityModel Find(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<ProgramActivityModel>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT date_time, researcher_id, program_id, activity_type, param1_data, param2_data FROM ProgramActivity WHERE log_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {

                        return new ProgramActivityModel(id,
                                                        rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                        rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                        rdr.GetInt32(rdr.GetOrdinal("program_id")),
                                                        rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data")));
                    }
                    return null;
                }
            });
        }

        public static List<ProgramActivityModel> FindAllByArtifactId(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<ProgramActivityModel>>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                var result = new List<ProgramActivityModel>();

                cmd.CommandText = "SELECT date_time, researcher_id, activity_type, param1_data, param2_data FROM ProgramActivity WHERE program_id = @Id ORDER BY date_time DESC";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while(rdr.Read())
                    {

                        result.Add(new ProgramActivityModel(id,
                                                        rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                        rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                        id,
                                                        rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data"))));
                    }
                }

                return result;
            });
        }

        public override ActivityModelKind GetKind() { return ActivityModelKind.Program; }
    }

    public class ResearchArticleActivityModel : ActivityModel
    {
        public const int CREATE = 1;
        public const int ARCHIVED = 2;
        public const int ADD_CITATION = 3;
        public const int REMOVE_CITATION = 4;
        public const int ADD_FILE = 5;
        public const int REMOVE_FILE = 6;
        public const int ADD_URL = 7;
        public const int REMOVE_URL = 8;

        public Lazy<ResearchArticleModel> ResearchArticle;

        public ResearchArticleActivityModel(int id, DateTime dt, int researcher, int research_article, int type, String param1_data, String param2_data)
            : base(id, dt, researcher, type, param1_data, param2_data)
        {
            this.ResearchArticle = new Lazy<ResearchArticleModel>(delegate() { return ResearchArticleModel.Find(research_article); }, research_article);
        }

        public static ResearchArticleActivityModel Create(DateTime dt, int researcher, int research_article, int type, String param1_data, String param2_data)
        {
            var result = DatabaseFacade.ExecuteTypedTx<ResearchArticleActivityModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO ResearchArticleActivity (date_time, researcher_id, research_article_id, activity_type, param1_data, param2_data) OUTPUT INSERTED.log_id VALUES (@DateTime, @ResearcherId, @Id, @ActivityType, @Param1Data, @Param2Data)";
                cmd.Parameters.Add("@DateTime", SqlDbType.DateTime).Value = dt;
                cmd.Parameters.Add("@ResearcherId", SqlDbType.Int).Value = researcher;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = research_article;
                cmd.Parameters.Add("@ActivityType", SqlDbType.Int).Value = type;
                cmd.Parameters.Add("@Param1Data", SqlDbType.NVarChar, 128).Value = (param1_data == null) ? SqlString.Null : param1_data;
                cmd.Parameters.Add("@Param2Data", SqlDbType.NVarChar, 128).Value = (param2_data == null) ? SqlString.Null : param2_data;
                cmd.Prepare();
                return new ResearchArticleActivityModel((Int32)cmd.ExecuteScalar(), dt, researcher, research_article, type, param1_data, param2_data);
            });

            var subs = new List<SubscriptionModel>();
            subs.AddRange(UserSubscriptionModel.FindAllSubscribers(researcher));
            subs.AddRange(ProgramSubscriptionModel.FindAllSubscribers((int)result.ResearchArticle.Value.Program.RawValue));

            foreach (var sub in subs)
            {
                int value = (int)sub.Subscriber.RawValue;
                if (value != researcher)
                {
                    ResearchArticleNotificationModel.Create(value, result.Id);
                }
            }

            return result;
        }

        public static ResearchArticleActivityModel CreateCreateActivity(DateTime dt, int researcher, int research_article)
        {
            return Create(dt, researcher, research_article, CREATE, null, null);
        }

        public static ResearchArticleActivityModel CreateArchivedActivity(DateTime dt, int researcher, int research_article)
        {
            return Create(dt, researcher, research_article, ARCHIVED, null, null);
        }

        public static ResearchArticleActivityModel CreateAddCitationActivity(DateTime dt, int researcher, int research_article, String citation)
        {
            return Create(dt, researcher, research_article, ADD_CITATION, citation, null);
        }

        public static ResearchArticleActivityModel CreateRemoveCitationActivity(DateTime dt, int researcher, int research_article, String citation)
        {
            return Create(dt, researcher, research_article, REMOVE_CITATION, citation, null);
        }

        public static ResearchArticleActivityModel CreateAddFileActivity(DateTime dt, int researcher, int research_article)
        {
            return Create(dt, researcher, research_article, ADD_FILE, null, null);
        }

        public static ResearchArticleActivityModel CreateRemoveFileActivity(DateTime dt, int researcher, int research_article)
        {
            return Create(dt, researcher, research_article, REMOVE_FILE, null, null);
        }

        public static ResearchArticleActivityModel CreateAddUrlActivity(DateTime dt, int researcher, int research_article)
        {
            return Create(dt, researcher, research_article, ADD_URL, null, null);
        }

        public static ResearchArticleActivityModel CreateRemoveUrlActivity(DateTime dt, int researcher, int research_article)
        {
            return Create(dt, researcher, research_article, REMOVE_URL, null, null);
        }

        public static ResearchArticleActivityModel Find(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<ResearchArticleActivityModel>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT date_time, researcher_id, research_article_id, activity_type, param1_data, param2_data FROM ResearchArticleActivity WHERE log_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new ResearchArticleActivityModel(id,
                                                                rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                                rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("research_article_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                                DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                                DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data")));
                    }
                    return null;
                }
            });
        }

        public static List<ResearchArticleActivityModel> FindAllByArtifactId(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<ResearchArticleActivityModel>>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                var result = new List<ResearchArticleActivityModel>();

                cmd.CommandText = "SELECT date_time, researcher_id, activity_type, param1_data, param2_data FROM ResearchArticleActivity WHERE research_article_id = @Id ORDER BY date_time DESC";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {

                        result.Add(new ResearchArticleActivityModel(id,
                                                        rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                        rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                        id,
                                                        rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data"))));
                    }
                }

                return result;
            });
        }

        public override ActivityModelKind GetKind() { return ActivityModelKind.ResearchArticle; }
    }

    public class JournalArticleActivityModel : ActivityModel
    {
        public const int CREATE = 1;
        public const int ARCHIVED = 2;
        public const int ADD_CITATION = 3;
        public const int REMOVE_CITATION = 4;
        public const int ADD_FILE = 5;
        public const int REMOVE_FILE = 6;
        public const int ADD_URL = 7;
        public const int REMOVE_URL = 8;

        public Lazy<JournalArticleModel> JournalArticle;

        public JournalArticleActivityModel(int id, DateTime dt, int researcher, int journal_article, int type, String param1_data, String param2_data)
            : base(id, dt, researcher, type, param1_data, param2_data)
        {
            this.JournalArticle = new Lazy<JournalArticleModel>(delegate() { return JournalArticleModel.Find(journal_article); }, journal_article);
        }

        public static JournalArticleActivityModel Create(DateTime dt, int researcher, int journal_article, int type, String param1_data, String param2_data)
        {
            var result = DatabaseFacade.ExecuteTypedTx<JournalArticleActivityModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO JournalArticleActivity (date_time, researcher_id, journal_article_id, activity_type, param1_data, param2_data) OUTPUT INSERTED.log_id VALUES (@DateTime, @ResearcherId, @Id, @ActivityType, @Param1Data, @Param2Data)";
                cmd.Parameters.Add("@DateTime", SqlDbType.DateTime).Value = dt;
                cmd.Parameters.Add("@ResearcherId", SqlDbType.Int).Value = researcher;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = journal_article;
                cmd.Parameters.Add("@ActivityType", SqlDbType.Int).Value = type;
                cmd.Parameters.Add("@Param1Data", SqlDbType.NVarChar, 128).Value = (param1_data == null) ? SqlString.Null : param1_data;
                cmd.Parameters.Add("@Param2Data", SqlDbType.NVarChar, 128).Value = (param2_data == null) ? SqlString.Null : param2_data;
                cmd.Prepare();
                return new JournalArticleActivityModel((Int32)cmd.ExecuteScalar(), dt, researcher, journal_article, type, param1_data, param2_data);
            });

            var subs = new List<SubscriptionModel>();
            subs.AddRange(UserSubscriptionModel.FindAllSubscribers(researcher));
            subs.AddRange(ProgramSubscriptionModel.FindAllSubscribers((int)result.JournalArticle.Value.Program.RawValue));

            foreach (var sub in subs)
            {
                int value = (int)sub.Subscriber.RawValue;
                if (value != researcher)
                {
                    JournalArticleNotificationModel.Create(value, result.Id);
                }
            }

            return result;
        }

        public static JournalArticleActivityModel CreateCreateActivity(DateTime dt, int researcher, int journal_article)
        {
            return Create(dt, researcher, journal_article, CREATE, null, null);
        }

        public static JournalArticleActivityModel CreateArchivedActivity(DateTime dt, int researcher, int journal_article)
        {
            return Create(dt, researcher, journal_article, ARCHIVED, null, null);
        }

        public static JournalArticleActivityModel CreateAddCitationActivity(DateTime dt, int researcher, int journal_article, String citation)
        {
            return Create(dt, researcher, journal_article, ADD_CITATION, citation, null);
        }

        public static JournalArticleActivityModel CreateRemoveCitationActivity(DateTime dt, int researcher, int journal_article, String citation)
        {
            return Create(dt, researcher, journal_article, REMOVE_CITATION, citation, null);
        }

        public static JournalArticleActivityModel CreateAddFileActivity(DateTime dt, int researcher, int journal_article)
        {
            return Create(dt, researcher, journal_article, ADD_FILE, null, null);
        }

        public static JournalArticleActivityModel CreateRemoveFileActivity(DateTime dt, int researcher, int journal_article)
        {
            return Create(dt, researcher, journal_article, REMOVE_FILE, null, null);
        }

        public static JournalArticleActivityModel CreateAddUrlActivity(DateTime dt, int researcher, int journal_article)
        {
            return Create(dt, researcher, journal_article, ADD_URL, null, null);
        }

        public static JournalArticleActivityModel CreateRemoveUrlActivity(DateTime dt, int researcher, int journal_article)
        {
            return Create(dt, researcher, journal_article, REMOVE_URL, null, null);
        }

        public static JournalArticleActivityModel Find(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<JournalArticleActivityModel>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT date_time, researcher_id, journal_article_id, activity_type, param1_data, param2_data FROM JournalArticleActivity WHERE log_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new JournalArticleActivityModel(id,
                                                               rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                               rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                               rdr.GetInt32(rdr.GetOrdinal("journal_article_id")),
                                                               rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                               DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                               DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data")));
                    }
                    return null;
                }
            });
        }

        public static List<JournalArticleActivityModel> FindAllByArtifactId(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<JournalArticleActivityModel>>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                var result = new List<JournalArticleActivityModel>();

                cmd.CommandText = "SELECT date_time, researcher_id, activity_type, param1_data, param2_data FROM JournalArticleActivity WHERE journal_article_id = @Id ORDER BY date_time DESC";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {

                        result.Add(new JournalArticleActivityModel(id,
                                                        rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                        rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                        id,
                                                        rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data"))));
                    }
                }

                return result;
            });
        }

        public override ActivityModelKind GetKind() { return ActivityModelKind.JournalArticle; }
    }

    public class MediaInterviewActivityModel : ActivityModel
    {
        public const int CREATE = 1;
        public const int ARCHIVED = 2;
        public const int ADD_CITATION = 3;
        public const int REMOVE_CITATION = 4;
        public const int ADD_FILE = 5;
        public const int REMOVE_FILE = 6;
        public const int ADD_URL = 7;
        public const int REMOVE_URL = 8;

        public Lazy<MediaInterviewModel> MediaInterview;

        public MediaInterviewActivityModel(int id, DateTime dt, int researcher, int media_interview, int type, String param1_data, String param2_data)
            : base(id, dt, researcher, type, param1_data, param2_data)
        {
            this.MediaInterview = new Lazy<MediaInterviewModel>(delegate() { return MediaInterviewModel.Find(media_interview); }, media_interview);
        }

        public static MediaInterviewActivityModel Create(DateTime dt, int researcher, int media_interview, int type, String param1_data, String param2_data)
        {
            var result = DatabaseFacade.ExecuteTypedTx<MediaInterviewActivityModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO MediaInterviewActivity (date_time, researcher_id, interview_id, activity_type, param1_data, param2_data) OUTPUT INSERTED.log_id VALUES (@DateTime, @ResearcherId, @Id, @ActivityType, @Param1Data, @Param2Data)";
                cmd.Parameters.Add("@DateTime", SqlDbType.DateTime).Value = dt;
                cmd.Parameters.Add("@ResearcherId", SqlDbType.Int).Value = researcher;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = media_interview;
                cmd.Parameters.Add("@ActivityType", SqlDbType.Int).Value = type;
                cmd.Parameters.Add("@Param1Data", SqlDbType.NVarChar, 128).Value = (param1_data == null) ? SqlString.Null : param1_data;
                cmd.Parameters.Add("@Param2Data", SqlDbType.NVarChar, 128).Value = (param2_data == null) ? SqlString.Null : param2_data;
                cmd.Prepare();
                return new MediaInterviewActivityModel((Int32)cmd.ExecuteScalar(), dt, researcher, media_interview, type, param1_data, param2_data);
            });

            var subs = new List<SubscriptionModel>();
            subs.AddRange(UserSubscriptionModel.FindAllSubscribers(researcher));
            subs.AddRange(ProgramSubscriptionModel.FindAllSubscribers((int)result.MediaInterview.Value.Program.RawValue));

            foreach (var sub in subs)
            {
                int value = (int)sub.Subscriber.RawValue;
                if (value != researcher)
                {
                    MediaInterviewNotificationModel.Create(value, result.Id);
                }
            }

            return result;
        }

        public static MediaInterviewActivityModel CreateCreateActivity(DateTime dt, int researcher, int media_interview)
        {
            return Create(dt, researcher, media_interview, CREATE, null, null);
        }

        public static MediaInterviewActivityModel CreateArchivedActivity(DateTime dt, int researcher, int media_interview)
        {
            return Create(dt, researcher, media_interview, ARCHIVED, null, null);
        }

        public static MediaInterviewActivityModel CreateAddCitationActivity(DateTime dt, int researcher, int media_interview, String citation)
        {
            return Create(dt, researcher, media_interview, ADD_CITATION, citation, null);
        }

        public static MediaInterviewActivityModel CreateRemoveCitationActivity(DateTime dt, int researcher, int media_interview, String citation)
        {
            return Create(dt, researcher, media_interview, REMOVE_CITATION, citation, null);
        }

        public static MediaInterviewActivityModel CreateAddFileActivity(DateTime dt, int researcher, int media_interview)
        {
            return Create(dt, researcher, media_interview, ADD_FILE, null, null);
        }

        public static MediaInterviewActivityModel CreateRemoveFileActivity(DateTime dt, int researcher, int media_interview)
        {
            return Create(dt, researcher, media_interview, REMOVE_FILE, null, null);
        }

        public static MediaInterviewActivityModel CreateAddUrlActivity(DateTime dt, int researcher, int media_interview)
        {
            return Create(dt, researcher, media_interview, ADD_URL, null, null);
        }

        public static MediaInterviewActivityModel CreateRemoveUrlActivity(DateTime dt, int researcher, int media_interview)
        {
            return Create(dt, researcher, media_interview, REMOVE_URL, null, null);
        }

        public static MediaInterviewActivityModel Find(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<MediaInterviewActivityModel>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT date_time, researcher_id, interview_id, activity_type, param1_data, param2_data FROM MediaInterviewActivity WHERE log_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new MediaInterviewActivityModel(id,
                                                               rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                               rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                               rdr.GetInt32(rdr.GetOrdinal("interview_id")),
                                                               rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                               DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                               DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data")));
                    }
                    return null;
                }
            });
        }

        public static List<MediaInterviewActivityModel> FindAllByArtifactId(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<MediaInterviewActivityModel>>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                var result = new List<MediaInterviewActivityModel>();

                cmd.CommandText = "SELECT date_time, researcher_id, activity_type, param1_data, param2_data FROM MediaInterviewActivity WHERE interview_id = @Id ORDER BY date_time DESC";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {

                        result.Add(new MediaInterviewActivityModel(id,
                                                        rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                        rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                        id,
                                                        rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data"))));
                    }
                }

                return result;
            });
        }

        public override ActivityModelKind GetKind() { return ActivityModelKind.MediaInterview; }
    }

    public class ConferencePresentationActivityModel : ActivityModel
    {
        public const int CREATE = 1;
        public const int ARCHIVED = 2;
        public const int ADD_CITATION = 3;
        public const int REMOVE_CITATION = 4;
        public const int ADD_FILE = 5;
        public const int REMOVE_FILE = 6;
        public const int ADD_URL = 7;
        public const int REMOVE_URL = 8;

        public Lazy<ConferencePresentationModel> ConferencePresentation;

        public ConferencePresentationActivityModel(int id, DateTime dt, int researcher, int conference_presentation, int type, String param1_data, String param2_data)
            : base(id, dt, researcher, type, param1_data, param2_data)
        {
            this.ConferencePresentation = new Lazy<ConferencePresentationModel>(delegate() { return ConferencePresentationModel.Find(conference_presentation); }, conference_presentation);
        }

        public static ConferencePresentationActivityModel Create(DateTime dt, int researcher, int conference_presentation, int type, String param1_data, String param2_data)
        {
            var result = DatabaseFacade.ExecuteTypedTx<ConferencePresentationActivityModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO ConferencePresentationActivity (date_time, researcher_id, presentation_id, activity_type, param1_data, param2_data) OUTPUT INSERTED.log_id VALUES (@DateTime, @ResearcherId, @Id, @ActivityType, @Param1Data, @Param2Data)";
                cmd.Parameters.Add("@DateTime", SqlDbType.DateTime).Value = dt;
                cmd.Parameters.Add("@ResearcherId", SqlDbType.Int).Value = researcher;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = conference_presentation;
                cmd.Parameters.Add("@ActivityType", SqlDbType.Int).Value = type;
                cmd.Parameters.Add("@Param1Data", SqlDbType.NVarChar, 128).Value = (param1_data == null) ? SqlString.Null : param1_data;
                cmd.Parameters.Add("@Param2Data", SqlDbType.NVarChar, 128).Value = (param2_data == null) ? SqlString.Null : param2_data;
                cmd.Prepare();
                return new ConferencePresentationActivityModel((Int32)cmd.ExecuteScalar(), dt, researcher, conference_presentation, type, param1_data, param2_data);
            });

            var subs = new List<SubscriptionModel>();
            subs.AddRange(UserSubscriptionModel.FindAllSubscribers(researcher));
            subs.AddRange(ProgramSubscriptionModel.FindAllSubscribers((int)result.ConferencePresentation.Value.Program.RawValue));

            foreach (var sub in subs)
            {
                int value = (int)sub.Subscriber.RawValue;
                if (value != researcher)
                {
                    ConferencePresentationNotificationModel.Create(value, result.Id);
                }
            }

            return result;
        }

        public static ConferencePresentationActivityModel CreateCreateActivity(DateTime dt, int researcher, int conference_presentation)
        {
            return Create(dt, researcher, conference_presentation, CREATE, null, null);
        }

        public static ConferencePresentationActivityModel CreateArchivedActivity(DateTime dt, int researcher, int conference_presentation)
        {
            return Create(dt, researcher, conference_presentation, ARCHIVED, null, null);
        }

        public static ConferencePresentationActivityModel CreateAddCitationActivity(DateTime dt, int researcher, int conference_presentation, String citation)
        {
            return Create(dt, researcher, conference_presentation, ADD_CITATION, citation, null);
        }

        public static ConferencePresentationActivityModel CreateRemoveCitationActivity(DateTime dt, int researcher, int conference_presentation, String citation)
        {
            return Create(dt, researcher, conference_presentation, REMOVE_CITATION, citation, null);
        }

        public static ConferencePresentationActivityModel CreateAddFileActivity(DateTime dt, int researcher, int conference_presentation)
        {
            return Create(dt, researcher, conference_presentation, ADD_FILE, null, null);
        }

        public static ConferencePresentationActivityModel CreateRemoveFileActivity(DateTime dt, int researcher, int conference_presentation)
        {
            return Create(dt, researcher, conference_presentation, REMOVE_FILE, null, null);
        }

        public static ConferencePresentationActivityModel CreateAddUrlActivity(DateTime dt, int researcher, int conference_presentation)
        {
            return Create(dt, researcher, conference_presentation, ADD_URL, null, null);
        }

        public static ConferencePresentationActivityModel CreateRemoveUrlActivity(DateTime dt, int researcher, int conference_presentation)
        {
            return Create(dt, researcher, conference_presentation, REMOVE_URL, null, null);
        }

        public static ConferencePresentationActivityModel Find(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<ConferencePresentationActivityModel>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT date_time, researcher_id, presentation_id, activity_type, param1_data, param2_data FROM ConferencePresentationActivity WHERE log_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new ConferencePresentationActivityModel(id,
                                                                       rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                                       rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                                       rdr.GetInt32(rdr.GetOrdinal("presentation_id")),
                                                                       rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                                       DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                                       DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data")));
                    }
                    return null;
                }
            });
        }

        public static List<ConferencePresentationActivityModel> FindAllByArtifactId(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<ConferencePresentationActivityModel>>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                var result = new List<ConferencePresentationActivityModel>();

                cmd.CommandText = "SELECT date_time, researcher_id, activity_type, param1_data, param2_data FROM ConferencePresentationActivity WHERE presentation_id = @Id ORDER BY date_time DESC";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {

                        result.Add(new ConferencePresentationActivityModel(id,
                                                        rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                        rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                        id,
                                                        rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data"))));
                    }
                }

                return result;
            });
        }

        public override ActivityModelKind GetKind() { return ActivityModelKind.ConferencePresentation; }
    }

    public class GrantActivityModel : ActivityModel
    {
        public const int CREATE = 1;
        public const int ARCHIVED = 2;
        public const int ADD_CITATION = 3;
        public const int REMOVE_CITATION = 4;
        public const int ADD_FILE = 5;
        public const int REMOVE_FILE = 6;
        public const int ADD_URL = 7;
        public const int REMOVE_URL = 8;

        public Lazy<GrantModel> Grant;

        public GrantActivityModel(int id, DateTime dt, int researcher, int grant_id, int type, String param1_data, String param2_data)
            : base(id, dt, researcher, type, param1_data, param2_data)
        {
            this.Grant = new Lazy<GrantModel>(delegate() { return GrantModel.Find(grant_id); }, grant_id);
        }

        public static GrantActivityModel Create(DateTime dt, int researcher, int grant_id, int type, String param1_data, String param2_data)
        {
            var result = DatabaseFacade.ExecuteTypedTx<GrantActivityModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO GrantActivity (date_time, researcher_id, grant_id, activity_type, param1_data, param2_data) OUTPUT INSERTED.log_id VALUES (@DateTime, @ResearcherId, @Id, @ActivityType, @Param1Data, @Param2Data)";
                cmd.Parameters.Add("@DateTime", SqlDbType.DateTime).Value = dt;
                cmd.Parameters.Add("@ResearcherId", SqlDbType.Int).Value = researcher;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = grant_id;
                cmd.Parameters.Add("@ActivityType", SqlDbType.Int).Value = type;
                cmd.Parameters.Add("@Param1Data", SqlDbType.NVarChar, 128).Value = (param1_data == null) ? SqlString.Null : param1_data;
                cmd.Parameters.Add("@Param2Data", SqlDbType.NVarChar, 128).Value = (param2_data == null) ? SqlString.Null : param2_data;
                cmd.Prepare();
                return new GrantActivityModel((Int32)cmd.ExecuteScalar(), dt, researcher, grant_id, type, param1_data, param2_data);
            });

            var subs = new List<SubscriptionModel>();
            subs.AddRange(UserSubscriptionModel.FindAllSubscribers(researcher));
            subs.AddRange(ProgramSubscriptionModel.FindAllSubscribers((int)result.Grant.Value.Program.RawValue));

            foreach (var sub in subs)
            {
                int value = (int)sub.Subscriber.RawValue;
                if (value != researcher)
                {
                    GrantNotificationModel.Create(value, result.Id);
                }
            }

            return result;
        }

        public static GrantActivityModel CreateCreateActivity(DateTime dt, int researcher, int grant_id)
        {
            return Create(dt, researcher, grant_id, CREATE, null, null);
        }

        public static GrantActivityModel CreateArchivedActivity(DateTime dt, int researcher, int grant_id)
        {
            return Create(dt, researcher, grant_id, ARCHIVED, null, null);
        }

        public static GrantActivityModel CreateAddCitationActivity(DateTime dt, int researcher, int grant_id, String citation)
        {
            return Create(dt, researcher, grant_id, ADD_CITATION, citation, null);
        }

        public static GrantActivityModel CreateRemoveCitationActivity(DateTime dt, int researcher, int grant_id, String citation)
        {
            return Create(dt, researcher, grant_id, REMOVE_CITATION, citation, null);
        }

        public static GrantActivityModel CreateAddFileActivity(DateTime dt, int researcher, int grant_id)
        {
            return Create(dt, researcher, grant_id, ADD_FILE, null, null);
        }

        public static GrantActivityModel CreateRemoveFileActivity(DateTime dt, int researcher, int grant_id)
        {
            return Create(dt, researcher, grant_id, REMOVE_FILE, null, null);
        }

        public static GrantActivityModel CreateAddUrlActivity(DateTime dt, int researcher, int grant_id)
        {
            return Create(dt, researcher, grant_id, ADD_URL, null, null);
        }

        public static GrantActivityModel CreateRemoveUrlActivity(DateTime dt, int researcher, int grant_id)
        {
            return Create(dt, researcher, grant_id, REMOVE_URL, null, null);
        }

        public static GrantActivityModel Find(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<GrantActivityModel>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT date_time, researcher_id, grant_id, activity_type, param1_data, param2_data FROM GrantActivity WHERE log_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new GrantActivityModel(id,
                                                      rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                      rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                      rdr.GetInt32(rdr.GetOrdinal("grant_id")),
                                                      rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                      DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                      DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data")));
                    }
                    return null;
                }
            });
        }

        public static List<GrantActivityModel> FindAllByArtifactId(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<GrantActivityModel>>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                var result = new List<GrantActivityModel>();

                cmd.CommandText = "SELECT date_time, researcher_id, activity_type, param1_data, param2_data FROM GrantActivity WHERE grant_id = @Id ORDER BY date_time DESC";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {

                        result.Add(new GrantActivityModel(id,
                                                        rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                        rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                        id,
                                                        rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data"))));
                    }
                }

                return result;
            });
        }

        public override ActivityModelKind GetKind() { return ActivityModelKind.Grant; }
    }

    public class OtherActivityModel : ActivityModel
    {
        public const int CREATE = 1;
        public const int ARCHIVED = 2;
        public const int ADD_CITATION = 3;
        public const int REMOVE_CITATION = 4;
        public const int ADD_FILE = 5;
        public const int REMOVE_FILE = 6;
        public const int ADD_URL = 7;
        public const int REMOVE_URL = 8;

        public Lazy<OtherModel> Other;

        public OtherActivityModel(int id, DateTime dt, int researcher, int other_id, int type, String param1_data, String param2_data)
            : base(id, dt, researcher, type, param1_data, param2_data)
        {
            this.Other = new Lazy<OtherModel>(delegate() { return OtherModel.Find(other_id); }, other_id);
        }

        public static OtherActivityModel Create(DateTime dt, int researcher, int other_id, int type, String param1_data, String param2_data)
        {
            var result = DatabaseFacade.ExecuteTypedTx<OtherActivityModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO OtherActivity (date_time, researcher_id, other_id, activity_type, param1_data, param2_data) OUTPUT INSERTED.log_id VALUES (@DateTime, @ResearcherId, @Id, @ActivityType, @Param1Data, @Param2Data)";
                cmd.Parameters.Add("@DateTime", SqlDbType.DateTime).Value = dt;
                cmd.Parameters.Add("@ResearcherId", SqlDbType.Int).Value = researcher;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = other_id;
                cmd.Parameters.Add("@ActivityType", SqlDbType.Int).Value = type;
                cmd.Parameters.Add("@Param1Data", SqlDbType.NVarChar, 128).Value = (param1_data == null) ? SqlString.Null : param1_data;
                cmd.Parameters.Add("@Param2Data", SqlDbType.NVarChar, 128).Value = (param2_data == null) ? SqlString.Null : param2_data;
                cmd.Prepare();
                return new OtherActivityModel((Int32)cmd.ExecuteScalar(), dt, researcher, other_id, type, param1_data, param2_data);
            });

            var subs = new List<SubscriptionModel>();
            subs.AddRange(UserSubscriptionModel.FindAllSubscribers(researcher));
            subs.AddRange(ProgramSubscriptionModel.FindAllSubscribers((int)result.Other.Value.Program.RawValue));

            foreach (var sub in subs)
            {
                int value = (int)sub.Subscriber.RawValue;
                if (value != researcher)
                {
                    OtherNotificationModel.Create(value, result.Id);
                }
            }

            return result;
        }

        public static OtherActivityModel CreateCreateActivity(DateTime dt, int researcher, int other_id)
        {
            return Create(dt, researcher, other_id, CREATE, null, null);
        }

        public static OtherActivityModel CreateArchivedActivity(DateTime dt, int researcher, int other_id)
        {
            return Create(dt, researcher, other_id, ARCHIVED, null, null);
        }

        public static OtherActivityModel CreateAddCitationActivity(DateTime dt, int researcher, int other_id, String citation)
        {
            return Create(dt, researcher, other_id, ADD_CITATION, citation, null);
        }

        public static OtherActivityModel CreateRemoveCitationActivity(DateTime dt, int researcher, int other_id, String citation)
        {
            return Create(dt, researcher, other_id, REMOVE_CITATION, citation, null);
        }

        public static OtherActivityModel CreateAddFileActivity(DateTime dt, int researcher, int other_id)
        {
            return Create(dt, researcher, other_id, ADD_FILE, null, null);
        }

        public static OtherActivityModel CreateRemoveFileActivity(DateTime dt, int researcher, int other_id)
        {
            return Create(dt, researcher, other_id, REMOVE_FILE, null, null);
        }

        public static OtherActivityModel CreateAddUrlActivity(DateTime dt, int researcher, int other_id)
        {
            return Create(dt, researcher, other_id, ADD_URL, null, null);
        }

        public static OtherActivityModel CreateRemoveUrlActivity(DateTime dt, int researcher, int other_id)
        {
            return Create(dt, researcher, other_id, REMOVE_URL, null, null);
        }

        public static OtherActivityModel Find(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<OtherActivityModel>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT date_time, researcher_id, other_id, activity_type, param1_data, param2_data FROM OtherActivity WHERE log_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new OtherActivityModel(id,
                                                      rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                      rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                      rdr.GetInt32(rdr.GetOrdinal("other_id")),
                                                      rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                      DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                      DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data")));
                    }
                    return null;
                }
            });
        }

        public static List<OtherActivityModel> FindAllByArtifactId(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<OtherActivityModel>>(delegate(SqlConnection conn, SqlTransaction trans, SqlCommand cmd)
            {
                var result = new List<OtherActivityModel>();

                cmd.CommandText = "SELECT date_time, researcher_id, activity_type, param1_data, param2_data FROM OtherActivity WHERE other_id = @Id ORDER BY date_time DESC";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {

                        result.Add(new OtherActivityModel(id,
                                                        rdr.GetDateTime(rdr.GetOrdinal("date_time")),
                                                        rdr.GetInt32(rdr.GetOrdinal("researcher_id")),
                                                        id,
                                                        rdr.GetInt32(rdr.GetOrdinal("activity_type")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param1_data")),
                                                        DatabaseFacade.StringOrNull(rdr, rdr.GetOrdinal("param2_data"))));
                    }
                }

                return result;
            });
        }

        public override ActivityModelKind GetKind() { return ActivityModelKind.Other; }
    }
}