using System;
using System.Data.SqlClient;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Data;

namespace P5T.Models
{
    public class AttachmentModel
    {
        public int Id;
        public string Name;
        public string Description;
        public Lazy<byte[]> Data;
        public int? PrevRevId;

        AttachmentModel(int id, string name, string description, byte[] data, int? prev_rev_id = null)
        {
            Id = id;
            Name = name;
            Description = description;
            Data = new Lazy<byte[]>(data);
            PrevRevId = prev_rev_id;
        }

        AttachmentModel(int id, string name, string description, int? prev_rev_id = null)
        {
            Id = id;
            Name = name;
            Description = description;
            PrevRevId = prev_rev_id;
            Data = new Lazy<byte[]>(delegate()
            {
                // Use the id to get the actual data from the DB and return it.
                return DatabaseFacade.ExecuteTypedTx<byte[]>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT attachment_data FROM Attachment WHERE attachment_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        if(rdr.Read())
                        {
                            return rdr.GetSqlBinary(rdr.GetOrdinal("attachment_data")).Value;
                        }
                    }

                    return null;
                });
            }, id);
        }

        public static AttachmentModel Create(string name, string description, System.Data.SqlTypes.SqlBinary data, int? prev_rev_id)
        {
            // TODO : Create a new attachment and return it.
            return DatabaseFacade.ExecuteTypedTx<AttachmentModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO Attachment (attachment_name, attachment_data, attachment_description, prev_rev_id) OUTPUT INSERTED.attachment_id VALUES (@Name, @Data, @Description, @PrevRevId)";

                cmd.Parameters.Add("@Name", SqlDbType.NVarChar, 128).Value = name;
                cmd.Parameters.Add("@Description", SqlDbType.NVarChar, 128).Value = description;
                cmd.Parameters.Add("@Data", SqlDbType.VarBinary, Int32.MaxValue).Value = data;

                if (prev_rev_id.HasValue)
                {
                    cmd.Parameters.Add("@PrevRevId", SqlDbType.Int).Value = prev_rev_id.Value;
                    cmd.Prepare();
                    return new AttachmentModel((Int32)cmd.ExecuteScalar(), name, description, data.Value, prev_rev_id.Value);
                }
                else
                {
                    cmd.Parameters.Add("@PrevRevId", SqlDbType.Int).Value = DBNull.Value;
                    cmd.Prepare();
                    return new AttachmentModel((Int32)cmd.ExecuteScalar(), name, description, data.Value, null);
                }
            });
        }

        public static AttachmentModel Find(int id)
        {
            // TODO : Find an attachment model and return it.
            return DatabaseFacade.ExecuteTypedTx<AttachmentModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT attachment_name, attachment_description, prev_rev_id FROM Attachment WHERE attachment_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        string attachment_name = rdr.GetString(rdr.GetOrdinal("attachment_name"));
                        string attachment_desc = rdr.GetString(rdr.GetOrdinal("attachment_description"));
                        int? prev_rev_id = null;
                        if (!rdr.IsDBNull(rdr.GetOrdinal("prev_rev_id")))
                        {
                            prev_rev_id = rdr.GetInt32(rdr.GetOrdinal("prev_rev_id"));
                        }
                        return new AttachmentModel(id, attachment_name, attachment_desc, prev_rev_id);
                    }
                }

                return null;
            });
        }

        public static IEnumerable<AttachmentModel> FindPrevRevisions(AttachmentModel model)
        {
            LinkedList<AttachmentModel> revs = new LinkedList<AttachmentModel>();

            revs.AddFirst(model);

            AttachmentModel current = model;
            while (current.PrevRevId.HasValue)
            {
                var attachment = Find(current.PrevRevId.Value);
                if (attachment != null)
                {
                    revs.AddFirst(attachment);
                    current = attachment;
                }
            }

            return revs;
        }

        public static void Update(AttachmentModel model)
        {
            // TODO : Update the given attachment model in the DB.
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                if (model.Data.IsLoaded())
                {
                    cmd.CommandText = "UPDATE Attachment SET attachment_name = @Name, attachment_description = @Desc, attachment_data = @Data, prev_rev_id = @PrevRevId WHERE attachment_id = @Id";
                    cmd.Parameters.Add("@Name", SqlDbType.NVarChar, 128).Value = model.Name;
                    cmd.Parameters.Add("@Desc", SqlDbType.NVarChar, 128).Value = model.Description;
                    cmd.Parameters.Add("@Data", SqlDbType.VarBinary).Value = model.Data.Value;
                    if (model.PrevRevId.HasValue)
                    {
                        cmd.Parameters.Add("@PrevRevId", SqlDbType.Int).Value = model.PrevRevId.Value;
                    }
                    else
                    {
                        cmd.Parameters.Add("@PrevRevId", SqlDbType.Int).Value = DBNull.Value;
                    }
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = model.Id;
                }
                else
                {
                    cmd.CommandText = "UPDATE Attachment SET attachment_name = @Name, attachment_description = @Desc, prev_rev_id = @PrevRevId WHERE attachment_id = @Id";
                    cmd.Parameters.Add("@Name", SqlDbType.NVarChar, 128).Value = model.Name;
                    cmd.Parameters.Add("@Desc", SqlDbType.NVarChar, 128).Value = model.Description;
                    if (model.PrevRevId.HasValue)
                    {
                        cmd.Parameters.Add("@PrevRevId", SqlDbType.Int).Value = model.PrevRevId.Value;
                    }
                    else
                    {
                        cmd.Parameters.Add("@PrevRevId", SqlDbType.Int).Value = DBNull.Value;
                    }
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = model.Id;
                }
                cmd.Prepare();

                return cmd.ExecuteNonQuery() > 0;
            });
        }

        public static void Delete(int id)
        {
            // TODO : Delete the attachment with the proper ID from the DB.
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM Attachment WHERE attachment_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() > 0;
            });
        }
    }
}