﻿using System;
using System.Data.SqlClient;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Data;

namespace P5T.Models
{
    public class AuthorModel
    {
        public int Id;
        public AffiliationModel PrimaryAffiliation;
        public LazyList<AffiliationModel> SecondaryAffiliations;
        public string FirstName;
        public string LastName;
        public char? MiddleInitial;

        public override string ToString()
        {
            if (MiddleInitial.HasValue)
            {
                return String.Format("{0} {1}. {2}", FirstName, MiddleInitial.Value, LastName);
            }
            else
            {
                return String.Format("{0} {1}", FirstName, LastName);
            }
        }

        public AuthorModel(int id, AffiliationModel primary_affiliation, string first_name, string last_name, char? middle_initial)
        {
            Id = id;
            PrimaryAffiliation = primary_affiliation;
            FirstName = first_name;
            LastName = last_name;
            MiddleInitial = middle_initial;

            SecondaryAffiliations = new LazyList<AffiliationModel>(delegate()
            {
                List<AffiliationModel> result = new List<AffiliationModel>();

                // Query the DB for the Secondary affiliations.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT affiliation_id FROM AuthorAffiliation WHERE author_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            var affiliation = AffiliationModel.Find(rdr.GetInt32(rdr.GetOrdinal("affiliation_id")));
                            if (affiliation != null)
                            {
                                result.Add(affiliation);
                            }
                        }
                    }

                    return true;
                });

                return result;
            }, id);
        }

        public static List<AuthorModel> List()
        {
            return DatabaseFacade.ExecuteTypedTx<List<AuthorModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                List<AuthorModel> result = new List<AuthorModel>();

                cmd.CommandText = "SELECT author_id, affiliation_primary, first_name, last_name, middle_init FROM Author";
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        var id = rdr.GetInt32(rdr.GetOrdinal("author_id"));
                        var primary_affiliation = rdr.GetInt32(rdr.GetOrdinal("affiliation_primary"));
                        var first_name = rdr.GetString(rdr.GetOrdinal("first_name"));
                        var last_name = rdr.GetString(rdr.GetOrdinal("last_name"));
                        char? middle_init = null;

                        if (!rdr.IsDBNull(rdr.GetOrdinal("middle_init")))
                        {
                            middle_init = rdr.GetString(rdr.GetOrdinal("middle_init")).ElementAt(0);
                        }

                        result.Add(new AuthorModel(id, AffiliationModel.Find(primary_affiliation), first_name, last_name, middle_init));
                    }
                }

                return result;
            });
        }

        public static AuthorModel Create(AffiliationModel primary_affiliation, string firstname, string lastname, char? middle_initial, List<AffiliationModel> secondary_affiliations)
        {
            // Create a new Author in the DB and return it.
            return DatabaseFacade.ExecuteTypedTx<AuthorModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                AuthorModel result = null;

                cmd.CommandText = "INSERT INTO Author (affiliation_primary, first_name, last_name, middle_init) OUTPUT INSERTED.author_id VALUES (@PrimaryAffiliation, @FirstName, @LastName, @MiddleInit)";
                cmd.Parameters.Add("@PrimaryAffiliation", SqlDbType.Int).Value = primary_affiliation.Id;
                cmd.Parameters.Add("@FirstName", SqlDbType.NVarChar, 128).Value = firstname;
                cmd.Parameters.Add("@LastName", SqlDbType.NVarChar, 128).Value = lastname;
                if (middle_initial.HasValue)
                {
                    cmd.Parameters.Add("@MiddleInit", SqlDbType.NChar, 1).Value = middle_initial.Value;
                    cmd.Prepare();
                    result = new AuthorModel((Int32)cmd.ExecuteScalar(), primary_affiliation, firstname, lastname, middle_initial.Value);
                }
                else
                {
                    cmd.Parameters.Add("@MiddleInit", SqlDbType.NChar, 1).Value = DBNull.Value;
                    cmd.Prepare();
                    result = new AuthorModel((Int32)cmd.ExecuteScalar(), primary_affiliation, firstname, lastname, null);
                }

                foreach (AffiliationModel aff in secondary_affiliations)
                {
                    using (cmd = conn.CreateCommand())
                    {
                        cmd.Transaction = tx;
                        cmd.CommandText = "INSERT INTO AuthorAffilation (affiliation_id, author_id) VALUES (@AffiliationId, @AuthorId)";
                        cmd.Parameters.Add("@AffiliationId", SqlDbType.Int).Value = aff.Id;
                        cmd.Parameters.Add("@AuthorId", SqlDbType.Int).Value = result.Id;
                        cmd.Prepare();
                        cmd.ExecuteNonQuery();
                    }
                }

                return result;
            });
        }

        public static AuthorModel Find(int id)
        {
            // Find an author model and return it.
            return DatabaseFacade.ExecuteTypedTx<AuthorModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT affiliation_primary, first_name, last_name, middle_init FROM Author WHERE author_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        var primary_affiliation = rdr.GetInt32(rdr.GetOrdinal("affiliation_primary"));
                        var first_name = rdr.GetString(rdr.GetOrdinal("first_name"));
                        var last_name = rdr.GetString(rdr.GetOrdinal("last_name"));
                        char? middle_init = null;

                        if (!rdr.IsDBNull(rdr.GetOrdinal("middle_init")))
                        {
                            middle_init = rdr.GetString(rdr.GetOrdinal("middle_init")).ElementAt(0);
                        }

                        return new AuthorModel(id, AffiliationModel.Find(primary_affiliation), first_name, last_name, middle_init);
                    }
                    else
                    {
                        return null;
                    }
                }
            });
        }

        public static AuthorModel FindOrCreate(string name)
        {
            string[] parts = name.Split(' ');
            string first_name;
            string last_name;
            char? middle_init = null;

            if (parts.Length == 3)
            {
                first_name = parts[0];
                middle_init = parts[1][0];
                last_name = parts[2];
            }
            else if (parts.Length == 2)
            {
                first_name = parts[0];
                last_name = parts[1];
            }
            else
            {
                return null;
            }

            return DatabaseFacade.ExecuteTypedTx<AuthorModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT author_id, affiliation_primary FROM Author WHERE first_name = @FirstName and last_name = @LastName and middle_init " + (middle_init.HasValue ? "= @MiddleInit" : "IS NULL");
                cmd.Parameters.Add("@FirstName", SqlDbType.NVarChar, 128).Value = first_name;
                cmd.Parameters.Add("@LastName", SqlDbType.NVarChar, 128).Value = last_name;
                if (middle_init.HasValue)
                {
                    cmd.Parameters.Add("@MiddleInit", SqlDbType.NChar, 1).Value = middle_init.Value;
                }
                cmd.Prepare();

                // First we determine if an author already exists
                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        var author_id = rdr.GetInt32(rdr.GetOrdinal("author_id"));
                        var affiliation_primary = rdr.GetInt32(rdr.GetOrdinal("affiliation_primary"));
                        return new AuthorModel(author_id, AffiliationModel.Find(affiliation_primary), first_name, last_name, middle_init);
                    }
                }

                return Create(AffiliationModel.Create("", "", ""), first_name, last_name, middle_init, new List<AffiliationModel>());
            });
        }

        public static void Update(AuthorModel model)
        {
            // Update the given AuthorModel in the DB.
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "UPDATE Author SET affiliation_primary = @AffiliationPrimary, first_name = @FirstName, last_name = @LastName, middle_init = @MiddleInit WHERE author_id = @Id";
                cmd.Parameters.Add("@AffiliationPrimary", SqlDbType.Int).Value = model.PrimaryAffiliation.Id;
                cmd.Parameters.Add("@FirstName", SqlDbType.NVarChar, 128).Value = model.FirstName;
                cmd.Parameters.Add("@LastName", SqlDbType.NVarChar, 128).Value = model.LastName;
                if (model.MiddleInitial.HasValue)
                {
                    cmd.Parameters.Add("@MiddleInit", SqlDbType.NChar, 1).Value = model.MiddleInitial.Value;
                }
                else
                {
                    cmd.Parameters.Add("@MiddleInit", SqlDbType.NChar, 1).Value = DBNull.Value;
                }
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = model.Id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() > 0;
            });
        }

        public static void Delete(int id)
        {
            // Delete the given AuthorModel from the DB.
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM Author WHERE author_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() > 0;
            });
        }
    }
}