using System;
using System.Data.SqlClient;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Data;

namespace P5T.Models
{
    public class GrantModel
    {
        public int Id;
        public Lazy<ProgramModel> Program;
        public string Description;
        public string GrantSource;
        public string GrantNumber;
        public string GrantTitle;
        public DateTime GrantStart;
        public DateTime GrantEnd;
        public DateTime? ArchivalDate;

        public LazyList<AuthorModel> Authors;
        public LazyList<AttachmentModel> Attachments;
        public LazyList<UrlModel> Urls;

        public GrantModel(int id, ProgramModel program, string description, string grant_source, string grant_number, string grant_title, DateTime start, DateTime end, DateTime? ArchivalDate)
        {
            Id = id;
            Program = new Lazy<ProgramModel>(program);
            Description = description;
            GrantSource = grant_source;
            GrantNumber = grant_number;
            GrantTitle = grant_title;
            GrantStart = start;
            GrantEnd = end;

            InitializeLazyLists();
        }

        public GrantModel(int id, int program_id, string description, string grant_source, string grant_number, string grant_title, DateTime start, DateTime end, DateTime? ArchivalDate)
        {
            Id = id;
            Description = description;
            GrantSource = grant_source;
            GrantNumber = grant_number;
            GrantTitle = grant_title;
            GrantStart = start;
            GrantEnd = end;

            Program = new Lazy<ProgramModel>(delegate()
            {
                return ProgramModel.Find(program_id);
            }, program_id);

            InitializeLazyLists();
        }

        public String GetMetadata()
        {
            String result = String.Format("Description: {0}\nGrant Source: {1}\nGrant Number: {2}\nGrant Title: {3}\nGrant Start: {4}\nGrant End: {5}\nAuthors:\n",
                Description, GrantSource, GrantNumber, GrantTitle, GrantStart, GrantEnd);
            foreach (var s in Authors)
            {
                result += "\t" + s.ToString();
            }
            return result;
        }

        private void InitializeLazyLists()
        {
            Authors = new LazyList<AuthorModel>(delegate()
            {
                var result = new List<AuthorModel>();
                // Use Id to get the authors from the database and return it.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT author_id FROM GrantAuthor WHERE grant_id = @GrantId";
                    cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = Id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            var author = AuthorModel.Find(rdr.GetInt32(rdr.GetOrdinal("author_id")));
                            if (author != null)
                            {
                                result.Add(author);
                            }
                        }
                    }

                    return true;
                });
                return result;
            }, Id);

            Attachments = new LazyList<AttachmentModel>(delegate()
            {
                var result = new List<AttachmentModel>();
                // Use Id to get the attachments from the DB and return it.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT attachment_id FROM GrantAttachment WHERE grant_id = @GrantId";
                    cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = Id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            var attachment = AttachmentModel.Find(rdr.GetInt32(rdr.GetOrdinal("attachment_id")));
                            if (attachment != null)
                            {
                                result.Add(attachment);
                            }
                        }
                    }
                    return true;
                });
                return result;
            }, Id);

            Urls = new LazyList<UrlModel>(delegate()
            {
                var result = new List<UrlModel>();
                // Use Id to get the Urls from the DB and return it.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT url_id FROM GrantUrl WHERE grant_id = @GrantId";
                    cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = Id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            var url = UrlModel.Find(rdr.GetInt32(rdr.GetOrdinal("url_id")));
                            if (url != null)
                            {
                                result.Add(url);
                            }
                        }
                    }
                    return true;
                });
                return result;
            }, Id);
        }

        public static GrantModel Create(ProgramModel program, string description, string grant_source, string grant_number, string grant_title, DateTime start, DateTime end)
        {
            // TODO : Create a new GrantModel and return it.
            return DatabaseFacade.ExecuteTypedTx<GrantModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO [Grant] (program_id, grant_description, grant_source, grant_number, grant_title, grant_start, grant_end) OUTPUT INSERTED.grant_id "
                    + "VALUES (@ProgramId, @GrantDesc, @GrantSource, @GrantNumber, @GrantTitle, @GrantStart, @GrantEnd)";
                cmd.Parameters.Add("@ProgramId", SqlDbType.Int).Value = program.Id;
                cmd.Parameters.Add("@GrantDesc", SqlDbType.NVarChar, 128).Value = description;
                cmd.Parameters.Add("@GrantSource", SqlDbType.NVarChar, 128).Value = grant_source;
                cmd.Parameters.Add("@GrantNumber", SqlDbType.NVarChar, 128).Value = grant_number;
                cmd.Parameters.Add("@GrantTitle", SqlDbType.NVarChar, 128).Value = grant_title;
                cmd.Parameters.Add("@GrantStart", SqlDbType.Date).Value = start;
                cmd.Parameters.Add("@GrantEnd", SqlDbType.Date).Value = end;
                cmd.Prepare();
                return new GrantModel((Int32)cmd.ExecuteScalar(), program, description, grant_source, grant_number, grant_title, start, end, null);
            });
        }

        public static GrantModel Find(int id)
        {
            // Find the GrantModel with the given ID and return it.
            return DatabaseFacade.ExecuteTypedTx<GrantModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT program_id, grant_description, grant_source, grant_number, grant_title, grant_start, grant_end, archival_date FROM [Grant] WHERE grant_id = @GrantId";
                cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new GrantModel(id, 
                            rdr.GetInt32(rdr.GetOrdinal("program_id")),
                            rdr.GetString(rdr.GetOrdinal("grant_description")),
                            rdr.GetString(rdr.GetOrdinal("grant_source")),
                            rdr.GetString(rdr.GetOrdinal("grant_number")),
                            rdr.GetString(rdr.GetOrdinal("grant_title")),
                            rdr.GetDateTime(rdr.GetOrdinal("grant_start")),
                            rdr.GetDateTime(rdr.GetOrdinal("grant_end")),
                            DatabaseFacade.DateTimeOrNull(rdr, rdr.GetOrdinal("archival_date")));
                    }
                    else
                    {
                        return null;
                    }
                }
            });
        }

        public static void Update(GrantModel model)
        {
            // TODO : Update only the fields in the Grant table associated with the given model.
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "UPDATE [Grant] SET " 
                    + "program_id = @ProgramId, grant_description = @GrantDesc, grant_source = @GrantSource, "
                    + "grant_number = @GrantNumber, grant_title = @GrantTitle, grant_start = @GrantStart, grant_end = @GrantEnd WHERE grant_id = @GrantId";
                cmd.Parameters.Add("@ProgramId", SqlDbType.Int).Value = model.Program.Value.Id;
                cmd.Parameters.Add("@GrantDesc", SqlDbType.NVarChar, 128).Value = model.Description;
                cmd.Parameters.Add("@GrantSource", SqlDbType.NVarChar, 128).Value = model.GrantSource;
                cmd.Parameters.Add("@GrantNumber", SqlDbType.NVarChar, 128).Value = model.GrantNumber;
                cmd.Parameters.Add("@GrantTitle", SqlDbType.NVarChar, 128).Value = model.GrantTitle;
                cmd.Parameters.Add("@GrantStart", SqlDbType.Date).Value = model.GrantStart;
                cmd.Parameters.Add("@GrantEnd", SqlDbType.Date).Value = model.GrantEnd;
                cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = model.Id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static void AddAuthor(GrantModel model, int author)
        {
            // Add the Author to the given GrantModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT COUNT(*) FROM GrantAuthor WHERE grant_id=@GrantId AND author_id=@AuthorId";
                cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@AuthorId", SqlDbType.Int).Value = author;
                cmd.Parameters.Add("@AuthorOrder", SqlDbType.Int).Value = 0;
                cmd.Prepare();

                if (0 == (Int32)cmd.ExecuteScalar())
                {
                    cmd.CommandText = "INSERT INTO GrantAuthor ([grant_id], [author_id], [author_order]) VALUES (@GrantId, @AuthorId, @AuthorOrder)";
                    cmd.Prepare();
                    return cmd.ExecuteNonQuery() == 1;
                }
                return false;
            });
        }

        public static void RemoveAuthor(GrantModel model, int author)
        {
            // TODO : Remove the Author from the given GrantModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM GrantAuthor WHERE grant_id = @GrantId AND author_id = @AuthorId";
                cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@AuthorId", SqlDbType.Int).Value = author;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() > 0;
            });
        }

        public static void AddAttachment(GrantModel model, int attachment)
        {
            // TODO : Add the Attachment to the given GrantModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO GrantAttachment ([grant_id], [attachment_id], [attachment_order]) VALUES (@GrantId, @AttachmentId, @AttachmentOrder)";
                cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@AttachmentId", SqlDbType.Int).Value = attachment;
                cmd.Parameters.Add("@AttachmentOrder", SqlDbType.Int).Value = 0;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static void UpdateAttachment(GrantModel model, int old_attachment, int new_attachment)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "UPDATE GrantAttachment SET attachment_id = @NewAttachmentId WHERE grant_id = @Id AND attachment_id = @OldAttachmentId";
                cmd.Parameters.Add("@NewAttachmentId", SqlDbType.Int).Value = new_attachment;
                cmd.Parameters.Add("@OldAttachmentId", SqlDbType.Int).Value = old_attachment;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = model.Id;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() > 0;
            });
        }

        public static void RemoveAttachment(GrantModel model, int attachment)
        {
            // TODO : Remove the Attachment from the given GrantModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM GrantAttachment WHERE grant_id = @GrantId AND attachment_id = @AttachmentId";
                cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@AttachmentId", SqlDbType.Int).Value = attachment;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() > 0;
            });
        }

        public static void AddUrl(GrantModel model, int url)
        {
            // TODO : Add the URL to the given GrantModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO GrantUrl ([grant_id], [url_id], [url_order]) VALUES (@GrantId, @UrlId, @UrlOrder)";
                cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@UrlId", SqlDbType.Int).Value = url;
                cmd.Parameters.Add("@UrlOrder", SqlDbType.Int).Value = 0;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static void RemoveUrl(GrantModel model, int url)
        {
            // TODO : Remove the URL from the given GrantModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM GrantUrl WHERE grant_id = @GrantId AND url_id = @UrlId";
                cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@UrlId", SqlDbType.Int).Value = url;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() > 0;
            });
        }

        public static void Delete(int id)
        { 
            // TODO : Delete the given grant from the DB.
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "UPDATE [Grant] SET archival_date = CONVERT(date, GETDATE()) WHERE grant_id = @GrantId";
                cmd.Parameters.Add("@GrantId", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static LazyList<GrantModel> FindByProgram(int program_id, bool archived = false)
        {
            return new LazyList<GrantModel>(delegate()
            {
                var result = new List<GrantModel>();

                // Get the list of grants from the DB.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT grant_id, grant_description, grant_source, grant_number, grant_title, grant_start, grant_end, archival_date FROM [Grant] WHERE program_id = @ProgramId AND archival_date IS " + (archived ? "NOT " : "") + "NULL";
                    cmd.Parameters.Add("@ProgramId", SqlDbType.Int).Value = program_id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            result.Add(new GrantModel(rdr.GetInt32(rdr.GetOrdinal("grant_id")),
                                            program_id,
                                            rdr.GetString(rdr.GetOrdinal("grant_description")),
                                            rdr.GetString(rdr.GetOrdinal("grant_source")),
                                            rdr.GetString(rdr.GetOrdinal("grant_number")),
                                            rdr.GetString(rdr.GetOrdinal("grant_title")),
                                            rdr.GetDateTime(rdr.GetOrdinal("grant_start")),
                                            rdr.GetDateTime(rdr.GetOrdinal("grant_end")),
                                            DatabaseFacade.DateTimeOrNull(rdr, rdr.GetOrdinal("archival_date"))));
                        }
                    }

                    return true;
                });

                return result;
            }, program_id);
        }

        public static LazyList<GrantModel> FindByAuthor(int id)
        {
            return new LazyList<GrantModel>(delegate()
            {
                var result = new List<GrantModel>();
                // Use Id to get the Journals from the DB and return them.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT grant_id FROM GrantAuthor WHERE author_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            var article = GrantModel.Find(rdr.GetInt32(rdr.GetOrdinal("grant_id")));
                            if (article != null)
                            {
                                result.Add(article);
                            }
                        }
                    }

                    return true;
                });

                return result;
            }, id);
        }
    }
}