using System;
using System.Data.SqlClient;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Data;

namespace P5T.Models
{
    public class MediaInterviewModel
    {
        public int Id;
        public Lazy<ProgramModel> Program;
        public string Description;
        public string Topic;
        public string Location;
        public string Outlet;
        public DateTime Date;
        public DateTime? ArchivalDate;

        public LazyList<AuthorModel> Authors;
        public LazyList<AttachmentModel> Attachments;
        public LazyList<UrlModel> Urls;

        public MediaInterviewModel(int id, ProgramModel program, string description, string topic, string location, string outlet, DateTime date, DateTime? ArchivalDate)
        {
            Id = id;
            Program = new Lazy<ProgramModel>(program);
            Description = description;
            Topic = topic;
            Location = location;
            Outlet = outlet;
            Date = date;

            InitializeLazyLists();
        }

        public MediaInterviewModel(int id, int program_id, string description, string topic, string location, string outlet, DateTime date, DateTime? ArchivalDate)
        {
            Id = id;
            Description = description;
            Topic = topic;
            Location = location;
            Outlet = outlet;
            Date = date;
            
            Program = new Lazy<ProgramModel>(delegate()
            {
                return ProgramModel.Find(program_id);
            }, program_id);

            InitializeLazyLists();
        }

        public String GetMetadata()
        {
            String result = String.Format("Description: {0}\nTopic: {1}\nLocation: {2}\nOutlet: {3}\nDate: {4}\nAuthors: \n",
                Description, Topic, Location, Outlet, Date);
            foreach (var s in Authors)
            {
                result += "\t" + s.ToString();
            }
            return result;
        }

        private void InitializeLazyLists()
        {
            Authors = new LazyList<AuthorModel>(delegate()
            {
                var result = new List<AuthorModel>();
                // Use Id to get the authors from the database and return it.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT author_id FROM MediaInterviewAuthor WHERE interview_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = Id;
                    cmd.Prepare();

                    using (var reader = cmd.ExecuteReader())
                    {
                        while(reader.Read())
                        {
                            var author = AuthorModel.Find(reader.GetInt32(reader.GetOrdinal("author_id")));
                            if (author != null)
                            {
                                result.Add(author);
                            }
                        }
                    }
                    return true;
                });
                return result;
            }, Id);

            Attachments = new LazyList<AttachmentModel>(delegate()
            {
                var result = new List<AttachmentModel>();
                // TODO : Use Id to get the attachments from the DB and return it.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT attachment_id FROM MediaInterviewAttachment WHERE interview_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = Id;
                    cmd.Prepare();

                    using (var reader = cmd.ExecuteReader())
                    {
                        while (reader.Read())
                        {
                            var attachment = AttachmentModel.Find(reader.GetInt32(reader.GetOrdinal("attachment_id")));
                            if (attachment != null)
                            {
                                result.Add(attachment);
                            }
                        }
                    }
                    return true;
                });
                return result;
            }, Id);

            Urls = new LazyList<UrlModel>(delegate()
            {
                var result = new List<UrlModel>();
                // TODO : Use Id to get the Urls from the DB and return it.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT url_id FROM MediaInterviewUrl WHERE interview_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = Id;
                    cmd.Prepare();

                    using (var reader = cmd.ExecuteReader())
                    {
                        while (reader.Read())
                        {
                            var url = UrlModel.Find(reader.GetInt32(reader.GetOrdinal("url_id")));
                            if(url != null)
                            {
                                result.Add(url);
                            }
                        }
                    }
                    return true;
                });
                return result;
            }, Id);
        }

        public static MediaInterviewModel Create(ProgramModel program, string description, string topic, string outlet, string location, DateTime date)
        {
            // TODO : Create a new MediaInterviewModel and return it.
            var result = DatabaseFacade.ExecuteTypedTx<MediaInterviewModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO MediaInterview ([program_id], [interview_description], [interview_topic], [interview_outlet], [interview_location], [interview_date]) OUTPUT INSERTED.interview_id VALUES (@ProgId, @Description, @Topic, @Outlet, @Location, @Date)";
                cmd.Parameters.Add("@ProgId", SqlDbType.Int).Value = program.Id;
                cmd.Parameters.Add("@Description", SqlDbType.NVarChar, 128).Value = description;
                cmd.Parameters.Add("@Topic", SqlDbType.NVarChar, 128).Value = topic;
                cmd.Parameters.Add("@Outlet", SqlDbType.NVarChar, 64).Value = outlet;
                cmd.Parameters.Add("@Location", SqlDbType.NVarChar, 128).Value = location;
                cmd.Parameters.Add("@Date", SqlDbType.DateTime).Value = date;
                cmd.Prepare();
                return new MediaInterviewModel((Int32)cmd.ExecuteScalar(), program.Id, description, topic, location, outlet, date, null);
            });

            return result;
        }

        public static MediaInterviewModel Find(int id)
        {
            //  Find the MediaInterviewModel with the given ID and return it.
            return DatabaseFacade.ExecuteTypedTx<MediaInterviewModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT program_id, interview_description, interview_topic, interview_outlet, interview_location, interview_date, archival_date FROM MediaInterview WHERE interview_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var reader = cmd.ExecuteReader())
                {
                    if (reader.Read())
                    {
                        return new MediaInterviewModel(id,
                            reader.GetInt32(reader.GetOrdinal("program_id")),
                            reader.GetString(reader.GetOrdinal("interview_description")),
                            reader.GetString(reader.GetOrdinal("interview_topic")),
                            reader.GetString(reader.GetOrdinal("interview_location")),
                            reader.GetString(reader.GetOrdinal("interview_outlet")),
                            reader.GetDateTime(reader.GetOrdinal("interview_date")),
                            DatabaseFacade.DateTimeOrNull(reader, reader.GetOrdinal("archival_date")));
                    }
                }

                return null;
            });
        }

        public static void Update(MediaInterviewModel model)
        {
            // Update only the fields in the MediaInterviewModel table associated with the given model.
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "UPDATE MediaInterview SET interview_topic = @Topic, interview_outlet = @Outlet, interview_location = @Location, interview_date = @Date WHERE interview_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@Topic", SqlDbType.NVarChar, 128).Value = model.Topic;
                cmd.Parameters.Add("@Outlet", SqlDbType.NVarChar, 64).Value = model.Outlet;
                cmd.Parameters.Add("@Location", SqlDbType.NVarChar, 128).Value = model.Location;
                cmd.Parameters.Add("@Date", SqlDbType.DateTime).Value = model.Date;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static void AddAuthor(MediaInterviewModel model, int author)
        {
            // Add the Author to the given MediaInterviewModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT COUNT(*) FROM MediaInterviewAuthor WHERE interview_id=@IntId AND author_id=@AuthId";
                cmd.Parameters.Add("@IntId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@AuthId", SqlDbType.Int).Value = author;
                cmd.Parameters.Add("@AuthOrder", SqlDbType.Int).Value = 0;
                cmd.Prepare();

                if (0 == (Int32)cmd.ExecuteScalar())
                {
                    cmd.CommandText = "INSERT INTO MediaInterviewAuthor ([interview_id], [author_id], [author_order]) VALUES (@IntId, @AuthId, @AuthOrder)";
                    cmd.Prepare();
                    return cmd.ExecuteNonQuery() == 1;
                }
                return false;
            });
        }

        public static void RemoveAuthor(MediaInterviewModel model, int author)
        {
            // Remove the Author from the given MediaInterviewModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM MediaInterviewAuthor WHERE interview_id = @ModelId AND author_id = @Id";
                cmd.Parameters.Add("@ModelId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = author;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static void AddAttachment(MediaInterviewModel model, int attachment)
        {
            // Add the Attachment to the given MediaInterviewModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO MediaInterviewAttachment ([interview_id], [attachment_id], [attachment_order]) VALUES (@IntId, @AttId, @AttOrder)";
                cmd.Parameters.Add("@IntId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@AttId", SqlDbType.Int).Value = attachment;
                cmd.Parameters.Add("@AttOrder", SqlDbType.Int).Value = 0;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static void UpdateAttachment(MediaInterviewModel model, int old_attachment, int new_attachment)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "UPDATE MediaInterviewAttachment SET attachment_id = @NewAttachmentId WHERE interview_id = @Id AND attachment_id = @OldAttachmentId";
                cmd.Parameters.Add("@NewAttachmentId", SqlDbType.Int).Value = new_attachment;
                cmd.Parameters.Add("@OldAttachmentId", SqlDbType.Int).Value = old_attachment;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = model.Id;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() > 0;
            });
        }

        public static void RemoveAttachment(MediaInterviewModel model, int attachment)
        {
            // Remove the Attachment from the given MediaInterviewModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM MediaInterviewAttachment WHERE attachment_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = attachment;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static void AddUrl(MediaInterviewModel model, int url)
        {
            // Add the URL to the given MediaInterviewModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO MediaInterviewUrl ([interview_id], [url_id], [url_order]) VALUES (@IntId, @UrlId, @UrlOrder)";
                cmd.Parameters.Add("@IntId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@UrlId", SqlDbType.Int).Value = url;
                cmd.Parameters.Add("@UrlOrder", SqlDbType.Int).Value = 0;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static void RemoveUrl(MediaInterviewModel model, int url)
        {
            // Remove the URL from the given MediaInterviewModel
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM MediaInterviewUrl WHERE url_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = url;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static void Delete(int id)
        {
            // Delete the given MediaInterviewModel from the DB.
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "UPDATE MediaInterview SET archival_date = CONVERT(date, GETDATE()) WHERE interview_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() == 1;
            });
        }

        public static LazyList<MediaInterviewModel> FindByProgram(int program_id, bool archived = false)
        {
            return new LazyList<MediaInterviewModel>(delegate()
            {
                var result = new List<MediaInterviewModel>();

                // TODO : Get the list of media interviews from the DB.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT interview_id, interview_description, interview_topic, interview_outlet, interview_location, interview_date, archival_date FROM MediaInterview WHERE program_id = @ProgramId AND archival_date IS " + (archived ? "NOT " : "") + "NULL";
                    cmd.Parameters.Add("@ProgramId", SqlDbType.Int).Value = program_id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            result.Add(new MediaInterviewModel(rdr.GetInt32(rdr.GetOrdinal("interview_id")),
                                            program_id,
                                            rdr.GetString(rdr.GetOrdinal("interview_description")),
                                            rdr.GetString(rdr.GetOrdinal("interview_topic")),
                                            rdr.GetString(rdr.GetOrdinal("interview_location")),
                                            rdr.GetString(rdr.GetOrdinal("interview_outlet")),
                                            rdr.GetSqlDateTime(rdr.GetOrdinal("interview_date")).Value,
                                            DatabaseFacade.DateTimeOrNull(rdr, rdr.GetOrdinal("archival_date"))));
                        }
                    }

                    return true;
                });

                return result;
            }, program_id);
        }

        public static LazyList<MediaInterviewModel> FindByAuthor(int id)
        {
            return new LazyList<MediaInterviewModel>(delegate()
            {
                var result = new List<MediaInterviewModel>();
                // Use Id to get the Journals from the DB and return them.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT interview_id FROM MediaInterviewAuthor WHERE author_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            var article = MediaInterviewModel.Find(rdr.GetInt32(rdr.GetOrdinal("interview_id")));
                            if (article != null)
                            {
                                result.Add(article);
                            }
                        }
                    }

                    return true;
                });

                return result;
            }, id);
        }
    }
}