﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.Linq;
using System.Web;

namespace P5T.Models
{
    public abstract class NotificationModel
    {
        public enum NotificationModelKind
        {
            Program,
            ResearchArticle,
            JournalArticle,
            ConferencePresentation,
            MediaInterview,
            Grant,
            Other
        }

        public int Id;
        public Lazy<ResearcherModel> Notifiee;

        public NotificationModel(int id, int notifiee)
        {
            this.Id = id;
            this.Notifiee = new Lazy<ResearcherModel>(delegate() { return ResearcherModel.Find(notifiee); }, notifiee);
        }

        public static void DeleteAllForUser(int user_id)
        {
            ProgramNotificationModel.DeleteForUser(user_id);
            ResearchArticleNotificationModel.DeleteForUser(user_id);
            JournalArticleNotificationModel.DeleteForUser(user_id);
            ConferencePresentationNotificationModel.DeleteForUser(user_id);
            MediaInterviewNotificationModel.DeleteForUser(user_id);
            GrantNotificationModel.DeleteForUser(user_id);
            OtherNotificationModel.DeleteForUser(user_id);
        }

        public static List<NotificationModel> FindAllNotifications(int user_id)
        {
            List<NotificationModel> result = new List<NotificationModel>();

            result.AddRange(ProgramNotificationModel.FindAllByUserId(user_id));
            result.AddRange(ResearchArticleNotificationModel.FindAllByUserId(user_id));
            result.AddRange(JournalArticleNotificationModel.FindAllByUserId(user_id));
            result.AddRange(ConferencePresentationNotificationModel.FindAllByUserId(user_id));
            result.AddRange(MediaInterviewNotificationModel.FindAllByUserId(user_id));
            result.AddRange(GrantNotificationModel.FindAllByUserId(user_id));
            result.AddRange(OtherNotificationModel.FindAllByUserId(user_id));

            result.Sort(delegate(NotificationModel lhs, NotificationModel rhs) { return rhs.GetDateTime().CompareTo(lhs.GetDateTime()); });

            return result;
        }

        public static string GetNotificationCount(int user_id)
        {
            var numNotifications = DatabaseFacade.ExecuteTypedTx<int>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                int result = 0;

                cmd.Parameters.Add("UserId", SqlDbType.Int).Value = user_id;

                cmd.CommandText = "SELECT COUNT(*) FROM ProgramNotification WHERE user_id = @UserId";
                cmd.Prepare();
                result += (Int32)cmd.ExecuteScalar();

                cmd.CommandText = "SELECT COUNT(*) FROM ResearchArticleNotification WHERE user_id = @UserId";
                cmd.Prepare();
                result += (Int32)cmd.ExecuteScalar();

                cmd.CommandText = "SELECT COUNT(*) FROM JournalArticleNotification WHERE user_id = @UserId";
                cmd.Prepare();
                result += (Int32)cmd.ExecuteScalar();

                cmd.CommandText = "SELECT COUNT(*) FROM MediaInterviewNotification WHERE user_id = @UserId";
                cmd.Prepare();
                result += (Int32)cmd.ExecuteScalar();

                cmd.CommandText = "SELECT COUNT(*) FROM ConferencePresentationNotification WHERE user_id = @UserId";
                cmd.Prepare();
                result += (Int32)cmd.ExecuteScalar();

                cmd.CommandText = "SELECT COUNT(*) FROM GrantNotification WHERE user_id = @UserId";
                cmd.Prepare();
                result += (Int32)cmd.ExecuteScalar();

                cmd.CommandText = "SELECT COUNT(*) FROM OtherNotification WHERE user_id = @UserId";
                cmd.Prepare();
                result += (Int32)cmd.ExecuteScalar();

                return result;
            });

            if (numNotifications != 0)
                return numNotifications.ToString();
            else
                return "";
        }

        public abstract DateTime GetDateTime();

        public abstract NotificationModelKind GetKind();
    }

    public class ProgramNotificationModel : NotificationModel
    {
        public Lazy<ProgramActivityModel> Activity;

        public ProgramNotificationModel(int id, int notifiee, int log_id)
            : base(id, notifiee)
        {
            this.Activity = new Lazy<ProgramActivityModel>(delegate() { return ProgramActivityModel.Find(log_id); }, log_id);
        }

        public static ProgramNotificationModel Create(int notifiee, int log_id)
        {
            return DatabaseFacade.ExecuteTypedTx<ProgramNotificationModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT notification_id FROM ProgramNotification WHERE user_id = @Notifiee AND log_id = @LogId";
                cmd.Parameters.Add("@Notifiee", SqlDbType.Int).Value = notifiee;
                cmd.Parameters.Add("@LogId", SqlDbType.Int).Value = log_id;
                cmd.Prepare();

                var id = cmd.ExecuteScalar();
                if (id == null)
                {
                    cmd.CommandText = "INSERT INTO ProgramNotification (user_id, log_id) OUTPUT INSERTED.notification_id VALUES (@Notifiee, @LogId)";
                    cmd.Prepare();

                    id = cmd.ExecuteScalar();
                }
                return new ProgramNotificationModel((Int32)id, notifiee, log_id);
            });
        }

        public static void Delete(int id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ProgramNotification WHERE notification_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static void DeleteForUser(int user_id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ProgramNotification WHERE user_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = user_id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static List<ProgramNotificationModel> FindAllByUserId(int user_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<ProgramNotificationModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<ProgramNotificationModel>();
                cmd.CommandText = "SELECT notification_id, user_id, log_id FROM ProgramNotification WHERE user_id = @UserId";
                cmd.Parameters.Add("UserId", SqlDbType.Int).Value = user_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new ProgramNotificationModel(rdr.GetInt32(rdr.GetOrdinal("notification_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("user_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("log_id"))));
                    }
                }

                return result;
            });
        }

        public override DateTime GetDateTime() { return Activity.Value.Dt; }

        public override NotificationModelKind GetKind() { return NotificationModelKind.Program; }
    }

    public class ResearchArticleNotificationModel : NotificationModel
    {
        public Lazy<ResearchArticleActivityModel> Activity;

        public ResearchArticleNotificationModel(int id, int notifiee, int log_id)
            : base(id, notifiee)
        {
            this.Activity = new Lazy<ResearchArticleActivityModel>(delegate() { return ResearchArticleActivityModel.Find(log_id); }, log_id);
        }

        public static ResearchArticleNotificationModel Create(int notifiee, int log_id)
        {
            return DatabaseFacade.ExecuteTypedTx<ResearchArticleNotificationModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT notification_id FROM ResearchArticleNotification WHERE user_id = @Notifiee AND log_id = @LogId";
                cmd.Parameters.Add("@Notifiee", SqlDbType.Int).Value = notifiee;
                cmd.Parameters.Add("@LogId", SqlDbType.Int).Value = log_id;
                cmd.Prepare();

                var id = cmd.ExecuteScalar();
                if (id == null)
                {
                    cmd.CommandText = "INSERT INTO ResearchArticleNotification (user_id, log_id) OUTPUT INSERTED.notification_id VALUES (@Notifiee, @LogId)";
                    cmd.Prepare();

                    id = cmd.ExecuteScalar();
                }
                return new ResearchArticleNotificationModel((Int32)id, notifiee, log_id);
            });
        }

        public static void Delete(int id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ResearchArticleNotification WHERE notification_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static void DeleteForUser(int user_id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ResearchArticleNotification WHERE user_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = user_id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static List<ResearchArticleNotificationModel> FindAllByUserId(int user_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<ResearchArticleNotificationModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<ResearchArticleNotificationModel>();
                cmd.CommandText = "SELECT notification_id, user_id, log_id FROM ResearchArticleNotification WHERE user_id = @UserId";
                cmd.Parameters.Add("UserId", SqlDbType.Int).Value = user_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new ResearchArticleNotificationModel(rdr.GetInt32(rdr.GetOrdinal("notification_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("user_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("log_id"))));
                    }
                }

                return result;
            });
        }

        public override DateTime GetDateTime() { return Activity.Value.Dt; }

        public override NotificationModelKind GetKind() { return NotificationModelKind.ResearchArticle; }
    }

    public class JournalArticleNotificationModel : NotificationModel
    {
        public Lazy<JournalArticleActivityModel> Activity;

        public JournalArticleNotificationModel(int id, int notifiee, int log_id)
            : base(id, notifiee)
        {
            this.Activity = new Lazy<JournalArticleActivityModel>(delegate() { return JournalArticleActivityModel.Find(log_id); }, log_id);
        }

        public static JournalArticleNotificationModel Create(int notifiee, int log_id)
        {
            return DatabaseFacade.ExecuteTypedTx<JournalArticleNotificationModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT notification_id FROM JournalArticleNotification WHERE user_id = @Notifiee AND log_id = @LogId";
                cmd.Parameters.Add("@Notifiee", SqlDbType.Int).Value = notifiee;
                cmd.Parameters.Add("@LogId", SqlDbType.Int).Value = log_id;
                cmd.Prepare();

                var id = cmd.ExecuteScalar();
                if (id == null)
                {
                    cmd.CommandText = "INSERT INTO JournalArticleNotification (user_id, log_id) OUTPUT INSERTED.notification_id VALUES (@Notifiee, @LogId)";
                    cmd.Prepare();

                    id = cmd.ExecuteScalar();
                }
                return new JournalArticleNotificationModel((Int32)id, notifiee, log_id);
            });
        }

        public static void Delete(int id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM JournalArticleNotification WHERE notification_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static void DeleteForUser(int user_id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM JournalArticleNotification WHERE user_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = user_id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static List<JournalArticleNotificationModel> FindAllByUserId(int user_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<JournalArticleNotificationModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<JournalArticleNotificationModel>();
                cmd.CommandText = "SELECT notification_id, user_id, log_id FROM JournalArticleNotification WHERE user_id = @UserId";
                cmd.Parameters.Add("UserId", SqlDbType.Int).Value = user_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new JournalArticleNotificationModel(rdr.GetInt32(rdr.GetOrdinal("notification_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("user_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("log_id"))));
                    }
                }

                return result;
            });
        }

        public override DateTime GetDateTime() { return Activity.Value.Dt; }

        public override NotificationModelKind GetKind() { return NotificationModelKind.JournalArticle; }
    }

    public class ConferencePresentationNotificationModel : NotificationModel
    {
        public Lazy<ConferencePresentationActivityModel> Activity;

        public ConferencePresentationNotificationModel(int id, int notifiee, int log_id)
            : base(id, notifiee)
        {
            this.Activity = new Lazy<ConferencePresentationActivityModel>(delegate() { return ConferencePresentationActivityModel.Find(log_id); }, log_id);
        }

        public static ConferencePresentationNotificationModel Create(int notifiee, int log_id)
        {
            return DatabaseFacade.ExecuteTypedTx<ConferencePresentationNotificationModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT notification_id FROM ConferencePresentationNotification WHERE user_id = @Notifiee AND log_id = @LogId";
                cmd.Parameters.Add("@Notifiee", SqlDbType.Int).Value = notifiee;
                cmd.Parameters.Add("@LogId", SqlDbType.Int).Value = log_id;
                cmd.Prepare();

                var id = cmd.ExecuteScalar();
                if (id == null)
                {
                    cmd.CommandText = "INSERT INTO ConferencePresentationNotification (user_id, log_id) OUTPUT INSERTED.notification_id VALUES (@Notifiee, @LogId)";
                    cmd.Prepare();

                    id = cmd.ExecuteScalar();
                }
                return new ConferencePresentationNotificationModel((Int32)id, notifiee, log_id);
            });
        }

        public static void Delete(int id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ConferencePresentationNotification WHERE notification_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static void DeleteForUser(int user_id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ConferencePresentationNotification WHERE user_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = user_id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static List<ConferencePresentationNotificationModel> FindAllByUserId(int user_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<ConferencePresentationNotificationModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<ConferencePresentationNotificationModel>();
                cmd.CommandText = "SELECT notification_id, user_id, log_id FROM ConferencePresentationNotification WHERE user_id = @UserId";
                cmd.Parameters.Add("UserId", SqlDbType.Int).Value = user_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new ConferencePresentationNotificationModel(rdr.GetInt32(rdr.GetOrdinal("notification_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("user_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("log_id"))));
                    }
                }

                return result;
            });
        }

        public override DateTime GetDateTime() { return Activity.Value.Dt; }

        public override NotificationModelKind GetKind() { return NotificationModelKind.ConferencePresentation; }
    }

    public class MediaInterviewNotificationModel : NotificationModel
    {
        public Lazy<MediaInterviewActivityModel> Activity;

        public MediaInterviewNotificationModel(int id, int notifiee, int log_id)
            : base(id, notifiee)
        {
            this.Activity = new Lazy<MediaInterviewActivityModel>(delegate() { return MediaInterviewActivityModel.Find(log_id); }, log_id);
        }

        public static MediaInterviewNotificationModel Create(int notifiee, int log_id)
        {
            return DatabaseFacade.ExecuteTypedTx<MediaInterviewNotificationModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT notification_id FROM MediaInterviewNotification WHERE user_id = @Notifiee AND log_id = @LogId";
                cmd.Parameters.Add("@Notifiee", SqlDbType.Int).Value = notifiee;
                cmd.Parameters.Add("@LogId", SqlDbType.Int).Value = log_id;
                cmd.Prepare();

                var id = cmd.ExecuteScalar();
                if (id == null)
                {
                    cmd.CommandText = "INSERT INTO MediaInterviewNotification (user_id, log_id) OUTPUT INSERTED.notification_id VALUES (@Notifiee, @LogId)";
                    cmd.Prepare();

                    id = cmd.ExecuteScalar();
                }
                return new MediaInterviewNotificationModel((Int32)id, notifiee, log_id);
            });
        }

        public static void Delete(int id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM MediaInterviewNotification WHERE notification_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static void DeleteForUser(int user_id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM MediaInterviewNotification WHERE user_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = user_id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static List<MediaInterviewNotificationModel> FindAllByUserId(int user_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<MediaInterviewNotificationModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<MediaInterviewNotificationModel>();
                cmd.CommandText = "SELECT notification_id, user_id, log_id FROM MediaInterviewNotification WHERE user_id = @UserId";
                cmd.Parameters.Add("UserId", SqlDbType.Int).Value = user_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new MediaInterviewNotificationModel(rdr.GetInt32(rdr.GetOrdinal("notification_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("user_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("log_id"))));
                    }
                }

                return result;
            });
        }

        public override DateTime GetDateTime() { return Activity.Value.Dt; }

        public override NotificationModelKind GetKind() { return NotificationModelKind.MediaInterview; }
    }

    public class GrantNotificationModel : NotificationModel
    {
        public Lazy<GrantActivityModel> Activity;

        public GrantNotificationModel(int id, int notifiee, int log_id)
            : base(id, notifiee)
        {
            this.Activity = new Lazy<GrantActivityModel>(delegate() { return GrantActivityModel.Find(log_id); }, log_id);
        }

        public static GrantNotificationModel Create(int notifiee, int log_id)
        {
            return DatabaseFacade.ExecuteTypedTx<GrantNotificationModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT notification_id FROM GrantNotification WHERE user_id = @Notifiee AND log_id = @LogId";
                cmd.Parameters.Add("@Notifiee", SqlDbType.Int).Value = notifiee;
                cmd.Parameters.Add("@LogId", SqlDbType.Int).Value = log_id;
                cmd.Prepare();

                var id = cmd.ExecuteScalar();
                if (id == null)
                {
                    cmd.CommandText = "INSERT INTO GrantNotification (user_id, log_id) OUTPUT INSERTED.notification_id VALUES (@Notifiee, @LogId)";
                    cmd.Prepare();

                    id = cmd.ExecuteScalar();
                }
                return new GrantNotificationModel((Int32)id, notifiee, log_id);
            });
        }

        public static void Delete(int id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM GrantNotification WHERE notification_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static void DeleteForUser(int user_id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM GrantNotification WHERE user_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = user_id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static List<GrantNotificationModel> FindAllByUserId(int user_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<GrantNotificationModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<GrantNotificationModel>();
                cmd.CommandText = "SELECT notification_id, user_id, log_id FROM GrantNotification WHERE user_id = @UserId";
                cmd.Parameters.Add("UserId", SqlDbType.Int).Value = user_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new GrantNotificationModel(rdr.GetInt32(rdr.GetOrdinal("notification_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("user_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("log_id"))));
                    }
                }

                return result;
            });
        }

        public override DateTime GetDateTime() { return Activity.Value.Dt; }

        public override NotificationModelKind GetKind() { return NotificationModelKind.Grant; }
    }

    public class OtherNotificationModel : NotificationModel
    {
        public Lazy<OtherActivityModel> Activity;

        public OtherNotificationModel(int id, int notifiee, int log_id)
            : base(id, notifiee)
        {
            this.Activity = new Lazy<OtherActivityModel>(delegate() { return OtherActivityModel.Find(log_id); }, log_id);
        }

        public static OtherNotificationModel Create(int notifiee, int log_id)
        {
            return DatabaseFacade.ExecuteTypedTx<OtherNotificationModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT notification_id FROM OtherNotification WHERE user_id = @Notifiee AND log_id = @LogId";
                cmd.Parameters.Add("@Notifiee", SqlDbType.Int).Value = notifiee;
                cmd.Parameters.Add("@LogId", SqlDbType.Int).Value = log_id;
                cmd.Prepare();

                var id = cmd.ExecuteScalar();
                if (id == null)
                {
                    cmd.CommandText = "INSERT INTO OtherNotification (user_id, log_id) OUTPUT INSERTED.notification_id VALUES (@Notifiee, @LogId)";
                    cmd.Prepare();

                    id = cmd.ExecuteScalar();
                }
                return new OtherNotificationModel((Int32)id, notifiee, log_id);
            });
        }

        public static void Delete(int id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM OtherNotification WHERE notification_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static void DeleteForUser(int user_id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM OtherNotification WHERE user_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = user_id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static List<OtherNotificationModel> FindAllByUserId(int user_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<OtherNotificationModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<OtherNotificationModel>();
                cmd.CommandText = "SELECT notification_id, user_id, log_id FROM OtherNotification WHERE user_id = @UserId";
                cmd.Parameters.Add("UserId", SqlDbType.Int).Value = user_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new OtherNotificationModel(rdr.GetInt32(rdr.GetOrdinal("notification_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("user_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("log_id"))));
                    }
                }

                return result;
            });
        }

        public override DateTime GetDateTime() { return Activity.Value.Dt; }

        public override NotificationModelKind GetKind() { return NotificationModelKind.Other; }
    }
}