using System;
using System.Data.SqlClient;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Data;

namespace P5T.Models
{
    public class ResearchArticleModel
    {
        public int Id;
        public Lazy<ProgramModel> Program;
        public string Description;
        public DateTime? ArchivalDate;

        public LazyList<AuthorModel> Authors;
        public LazyList<AttachmentModel> Attachments;
        public LazyList<UrlModel> Urls;

        public ResearchArticleModel(int id, ProgramModel program, string description, DateTime? ArchivalDate)
        {
            Id = id;
            Program = new Lazy<ProgramModel>(program);
            Description = description;

            InitializeLazyLists();
        }

        public ResearchArticleModel(int id, int program_id, string description, DateTime? ArchivalDate)
        {
            Id = id;
            Description = description;
            Program = new Lazy<ProgramModel>(delegate()
            {
                return ProgramModel.Find(program_id);
            }, program_id);

            InitializeLazyLists();
        }

        public String GetMetadata()
        {
            String result = String.Format("Description: {0}\nAuthors:\n", Description);
            foreach (var s in Authors)
            {
                result += "\t" + s.ToString();
            }
            return result;
        }

        private void InitializeLazyLists()
        {
            Authors = new LazyList<AuthorModel>(delegate()
            {
                var result = new List<AuthorModel>();
                // Use Id to get the authors from the database and return it.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT author_id FROM ResearchArticleAuthor WHERE research_article_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = Id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            var author = AuthorModel.Find(rdr.GetInt32(rdr.GetOrdinal("author_id")));
                            if (author != null)
                            {
                                result.Add(author);
                            }
                        }
                    }
                    return true;
                });
                return result;
            }, Id);

            Attachments = new LazyList<AttachmentModel>(delegate()
            {
                var result = new List<AttachmentModel>();
                // Use Id to get the attachments from the DB and return it.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT attachment_id FROM ResearchArticleAttachment WHERE research_article_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = Id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while(rdr.Read())
                        {
                            var attachment = AttachmentModel.Find(rdr.GetInt32(rdr.GetOrdinal("attachment_id")));
                            if (attachment != null)
                            {
                                result.Add(attachment);
                            }
                        }
                    }
                    return true;
                });
                return result;
            }, Id);

            Urls = new LazyList<UrlModel>(delegate()
            {
                var result = new List<UrlModel>();
                // Use Id to get the Urls from the DB and return it.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT url_id FROM ResearchArticleURL WHERE research_article_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = Id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while(rdr.Read())
                        {
                            var url = UrlModel.Find(rdr.GetInt32(rdr.GetOrdinal("url_id")));
                            if (url != null)
                            {
                                result.Add(url);
                            }
                        }
                    }
                    return true;
                });
                return result;
            }, Id);
        }

        public static ResearchArticleModel Create(ProgramModel program, string description)
        {
            // Create a new ResearchArticleModel and return it.
            return DatabaseFacade.ExecuteTypedTx<ResearchArticleModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO ResearchArticle ([program_id], [research_article_description]) OUTPUT INSERTED.research_article_id VALUES (@ProgId, @Description)";
                cmd.Parameters.Add("@ProgId", SqlDbType.Int).Value = program.Id;
                cmd.Parameters.Add("@Description", SqlDbType.NVarChar, 128).Value = description;
                cmd.Prepare();
                return new ResearchArticleModel((Int32)cmd.ExecuteScalar(), program, description, null);
            });
        }

        public static ResearchArticleModel Find(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<ResearchArticleModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT program_id, research_article_description, archival_date FROM ResearchArticle WHERE research_article_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new ResearchArticleModel(id,
                            rdr.GetInt32(rdr.GetOrdinal("program_id")),
                            rdr.GetString(rdr.GetOrdinal("research_article_description")),
                            DatabaseFacade.DateTimeOrNull(rdr, rdr.GetOrdinal("archival_date")));
                    }
                }
                return null;
            });
        }

        public static void Update(ResearchArticleModel model)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "UPDATE ResearchArticle SET research_article_description = @Description WHERE research_article_id = @Id";
                cmd.Parameters.Add("@Description", SqlDbType.NVarChar, 128).Value = model.Description;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = model.Id;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() >= 1;
            });
        }

        public static void AddAuthor(ResearchArticleModel model, int author)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT COUNT(*) FROM ResearchArticleAuthor WHERE research_article_id=@ArticleId AND author_id=@AuthId";
                cmd.Parameters.Add("@ArticleId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@AuthId", SqlDbType.Int).Value = author;
                cmd.Parameters.Add("@AuthOrder", SqlDbType.Int).Value = 0;
                cmd.Prepare();

                if (0 == (Int32)cmd.ExecuteScalar())
                {
                    cmd.CommandText = "INSERT INTO ResearchArticleAuthor ([research_article_id], [author_id], [author_order]) OUTPUT INSERTED.research_article_id VALUES(@ArticleId, @AuthId, @AuthOrder)";
                    cmd.Prepare();
                    return cmd.ExecuteNonQuery() >= 1;
                }
                return false;
            });
        }

        public static void RemoveAuthor(ResearchArticleModel model, int author)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ResearchArticleAuthor WHERE research_article_id = @ModelId AND author_id = @Id";
                cmd.Parameters.Add("@ModelId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = author;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() >= 1;
            });
        }

        public static void AddAttachment(ResearchArticleModel model, int attachment)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO ResearchArticleAttachment ([research_article_id], [attachment_id], [attachment_order]) OUTPUT INSERTED.research_article_id VALUES(@ArticleID, @AttId, @AttOrder)";
                cmd.Parameters.Add("@ArticleId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@AttId", SqlDbType.Int).Value = attachment;
                cmd.Parameters.Add("@AttOrder", SqlDbType.Int).Value = 0;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() >= 1;
            });
        }

        public static void UpdateAttachment(ResearchArticleModel model, int old_attachment, int new_attachment)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "UPDATE ResearchArticleAttachment SET attachment_id = @NewAttachmentId WHERE research_article_id = @Id AND attachment_id = @OldAttachmentId";
                cmd.Parameters.Add("@NewAttachmentId", SqlDbType.Int).Value = new_attachment;
                cmd.Parameters.Add("@OldAttachmentId", SqlDbType.Int).Value = old_attachment;
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = model.Id;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() > 0;
            });
        }

        public static void RemoveAttachment(ResearchArticleModel model, int attachment)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ResearchArticleAttachment WHERE attachment_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = attachment;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() >= 1;
            });
        }

        public static void AddUrl(ResearchArticleModel model, int url)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO ResearchArticleUrl ([research_article_id], [url_id], [url_order]) OUTPUT INSERTED.research_article_id VALUES (@ArticleId, @UrlId, @UrlOrder)";
                cmd.Parameters.Add("@ArticleId", SqlDbType.Int).Value = model.Id;
                cmd.Parameters.Add("@UrlId", SqlDbType.Int).Value = url;
                cmd.Parameters.Add("@UrlOrder", SqlDbType.Int).Value = 0;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() >= 1;
            });
        }

        public static void RemoveUrl(ResearchArticleModel model, int url)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ResearchArticleUrl WHERE url_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = url;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() >= 1;
            });
        }

        public static void Delete(int id)
        {
            // TODO : Delete the given ResearchArticleModel from the DB.
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "UPDATE ResearchArticle SET archival_date = CONVERT(date, GETDATE()) WHERE research_article_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                return cmd.ExecuteNonQuery() >= 1;
            });
        }

        public static LazyList<ResearchArticleModel> FindByProgram(int program_id, bool archived = false)
        {
            return new LazyList<ResearchArticleModel>(delegate()
            {
                var result = new List<ResearchArticleModel>();

                // Get the list of Research Articles from the DB.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT research_article_id, research_article_description, archival_date FROM ResearchArticle WHERE program_id = @ProgramId AND archival_date IS " + (archived ? "NOT " : "") + "NULL";
                    cmd.Parameters.Add("@ProgramId", SqlDbType.Int).Value = program_id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            result.Add(new ResearchArticleModel(
                                                rdr.GetInt32(rdr.GetOrdinal("research_article_id")),
                                                program_id, 
                                                rdr.GetString(rdr.GetOrdinal("research_article_description")),
                                                DatabaseFacade.DateTimeOrNull(rdr, rdr.GetOrdinal("archival_date"))));
                        }
                    }

                    return true;
                });

                return result;
            }, program_id);
        }

        public static LazyList<ResearchArticleModel> FindByAuthor(int id)
        {
            return new LazyList<ResearchArticleModel>(delegate()
            {
                var result = new List<ResearchArticleModel>();
                // Use Id to get the Journals from the DB and return them.
                DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
                {
                    cmd.CommandText = "SELECT research_article_id FROM ResearchArticleAuthor WHERE author_id = @Id";
                    cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                    cmd.Prepare();

                    using (var rdr = cmd.ExecuteReader())
                    {
                        while (rdr.Read())
                        {
                            var article = ResearchArticleModel.Find(rdr.GetInt32(rdr.GetOrdinal("research_article_id")));
                            if (article != null)
                            {
                                result.Add(article);
                            }
                        }
                    }

                    return true;
                });

                return result;
            }, id);
        }
    }
}