﻿using System;
using System.Data.SqlClient;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Mvc;
using System.Text;
using System.Security.Cryptography;
using System.Web.Configuration;
using System.Globalization;
using System.Data;

namespace P5T.Models
{
    public class ResearcherModel
    {
        public int Id;
        public Lazy<AuthorModel> Author;
        public string Username;
        public DateTime CreationTime;

        public ResearcherModel(int id, int author_id, string username, DateTime creationTime)
        {
            Id = id;
            Username = username;
            CreationTime = creationTime;
            Author = new Lazy<AuthorModel>(delegate()
            {
                // Use the DB to lookup the proper author by id and return it.
                return DatabaseFacade.ExecuteTypedTx<AuthorModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd) {
                    return AuthorModel.Find(author_id);
                });
            }, author_id);
        }

        public ResearcherModel(int id, AuthorModel author, string username, DateTime creationTime)
        {
            Id = id;
            Username = username;
            CreationTime = creationTime;
            Author = new Lazy<AuthorModel>(author);
        }

        public static ResearcherModel Create(AuthorModel author, string username, string password, DateTime creationTime)
        {
            var salt = GenerateSalt();
            var hash = GenerateHash(password, salt);

            // Create a new ResearcherModel and add it to the DB.
            return DatabaseFacade.ExecuteTypedTx<ResearcherModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "INSERT INTO Researcher (username, passhash, salt, creation_datetime, author_id) OUTPUT INSERTED.user_id VALUES (@Username, @Passhash, @Salt, @CreationDatetime, @AuthorId)";
                cmd.Parameters.Add("@Username", SqlDbType.NVarChar, 32).Value = username;
                cmd.Parameters.Add("@Passhash", SqlDbType.NChar, 128).Value = hash;
                cmd.Parameters.Add("@Salt", SqlDbType.NChar, 128).Value = BytesToHexString(salt);
                cmd.Parameters.Add("@CreationDatetime", SqlDbType.DateTime).Value = creationTime;
                cmd.Parameters.Add("@AuthorId", SqlDbType.Int).Value = author.Id;
                cmd.Prepare();

                return new ResearcherModel((Int32)cmd.ExecuteScalar(), author, username, creationTime);
            });
        }

        public static List<ResearcherModel> FindAll()
        {
            var result = new List<ResearcherModel>();

            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT user_id, username, author_id, creation_datetime FROM Researcher";
                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new ResearcherModel(rdr.GetInt32(rdr.GetOrdinal("user_id")),
                            rdr.GetInt32(rdr.GetOrdinal("author_id")),
                            rdr.GetString(rdr.GetOrdinal("username")),
                            rdr.GetSqlDateTime(rdr.GetOrdinal("creation_datetime")).Value));
                    }
                }

                return true;
            });

            return result;
        }

        public static ResearcherModel Find(int id)
        {
            // Find a Researcher in the DB and return it.
            return DatabaseFacade.ExecuteTypedTx<ResearcherModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT username, author_id, creation_datetime FROM Researcher WHERE user_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new ResearcherModel(id, 
                            rdr.GetInt32(rdr.GetOrdinal("author_id")), 
                            rdr.GetString(rdr.GetOrdinal("username")), 
                            rdr.GetSqlDateTime(rdr.GetOrdinal("creation_datetime")).Value);
                    }
                    else
                    {
                        return null;
                    }
                }
            });
        }

        public static ResearcherModel FindByAuthorId(int id)
        {
            return DatabaseFacade.ExecuteTypedTx<ResearcherModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT user_id, username, creation_datetime FROM Researcher WHERE author_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new ResearcherModel(rdr.GetInt32(rdr.GetOrdinal("user_id")),
                            id,
                            rdr.GetString(rdr.GetOrdinal("username")),
                            rdr.GetDateTime(rdr.GetOrdinal("creation_datetime")));
                    }
                    return null;
                }
            });
        }

        public static ResearcherModel Find(string username)
        {
            // Find a Researcher in the DB by username.
            return DatabaseFacade.ExecuteTypedTx<ResearcherModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT user_id, author_id, creation_datetime FROM Researcher where username = @Username";
                cmd.Parameters.Add("@Username", SqlDbType.NVarChar, 32).Value = username;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        return new ResearcherModel(rdr.GetInt32(rdr.GetOrdinal("user_id")),
                            rdr.GetInt32(rdr.GetOrdinal("author_id")),
                            username,
                            rdr.GetSqlDateTime(rdr.GetOrdinal("creation_datetime")).Value);
                    }
                    else
                    {
                        return null;
                    }
                }
            });
        }

        public static void ChangePassword(ResearcherModel model, string newpassword)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT salt FROM Researcher WHERE researcher_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = model.Id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    if (rdr.Read())
                    {
                        var salt = HexStringToBytes(rdr.GetString(rdr.GetOrdinal("salt")));
                        using (cmd = conn.CreateCommand())
                        {
                            cmd.Transaction = tx;
                            cmd.CommandText = "UPDATE Researcher SET passhash = @Passhash WHERE researcher_id = @Id";
                            cmd.Parameters.Add("@Id", SqlDbType.Int).Value = model.Id;
                            cmd.Parameters.Add("@Passhash", SqlDbType.NChar, 128).Value = GenerateHash(newpassword, salt);
                            cmd.Prepare();
                            return cmd.ExecuteNonQuery() > 0;
                        }
                    }
                    else
                    {
                        return false;
                    }
                }
            });
        }

        public static void Delete(int id)
        {
            // Delete a researcher from the DB.
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM Researcher WHERE user_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();

                return true;
            });
        }

        private static int HexCharToByte(char c)
        {
            if (c >= '0' && c <= '9') return (c - '0');
            if (c >= 'a' && c <= 'f') return (c - 'a') + 10;
            if (c >= 'A' && c <= 'F') return (c - 'A') + 10;
            throw new SystemException("Invalid character provided to HexCharToByte");
        }

        private static char[] lookup = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
        private static string BytesToHexString(byte[] data)
        {
            StringBuilder sb = new StringBuilder();
            foreach (byte b in data)
            {
                sb.Append(lookup[b >> 4]);
                sb.Append(lookup[b & 15]);
            }
            return sb.ToString();
        }

        private static byte[] HexStringToBytes(string hex)
        {
            int i = 0, j = 0; // Iteration Variables

            int offset = hex.Length % 2;
            byte[] result = new byte[hex.Length / 2 + offset];

            if (offset != 0)
            {
                result[0] = (byte)HexCharToByte(hex[0]);
                i = 1;
                j = 1;
            }

            for (; i < hex.Length; i += 2, ++j)
            {
                int a = HexCharToByte(hex[i]);
                int b = HexCharToByte(hex[i + 1]);

                result[j] = (byte)((HexCharToByte(hex[i]) << 4) + HexCharToByte(hex[i + 1]));
            }

            return result;
        }

        private static byte[] GenerateSalt()
        {
            Random random = new Random(unchecked((int)DateTime.Now.Ticks));

            if (random != null)
            {
                byte[] saltValue = new byte[64];
                random.NextBytes(saltValue);
                return saltValue;
            }

            return null;
        }

        private static string GenerateHash(string data, byte[] salt)
        {
            HashAlgorithm hasher = new HMACSHA512(HexStringToBytes(WebConfigurationManager.AppSettings["HMACKey"]));

            UnicodeEncoding encoding = new UnicodeEncoding();

            if (encoding != null && data != null && salt != null)
            {
                byte[] valueToHash = new byte[64 + encoding.GetByteCount(data)];

                salt.CopyTo(valueToHash, 0);
                encoding.GetBytes(data).CopyTo(valueToHash, 64);

                string hashedPassword = "";
                foreach (byte digit in hasher.ComputeHash(valueToHash))
                {
                    hashedPassword += digit.ToString("X2", CultureInfo.InvariantCulture.NumberFormat);
                }

                return hashedPassword;
            }

            return null;
        }

        public static bool Authenticate(string username, string password)
        {
            bool result = false;
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT passhash, salt FROM Researcher WHERE username = @Username";
                cmd.Parameters.Add("@Username", SqlDbType.NVarChar, 32).Value = username;
                cmd.Prepare();
                using (var reader = cmd.ExecuteReader())
                {
                    if (reader.Read())
                    {
                        var storedHash = reader.GetString(reader.GetOrdinal("passhash"));
                        var storedSalt = reader.GetString(reader.GetOrdinal("salt"));
                        var genHash = GenerateHash(password, HexStringToBytes(storedSalt));
                        result = storedHash == genHash;
                        return true;
                    }
                    else
                    {
                        return true;
                    }
                }
            });
            return result;
        }
    }
}
