﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.Linq;
using System.Web;

namespace P5T.Models
{
    public abstract class SubscriptionModel
    {
        public enum SubscriptionKind
        {
            User,
            Program
        }

        public int Id;
        public Lazy<ResearcherModel> Subscriber;

        public SubscriptionModel(int id, int subscriber)
        {
            this.Id = id;
            this.Subscriber = new Lazy<ResearcherModel>(delegate() { return ResearcherModel.Find(subscriber); }, subscriber);
        }

        public abstract SubscriptionKind GetKind();
    }

    public class UserSubscriptionModel : SubscriptionModel
    {
        public Lazy<ResearcherModel> Subscribee;

        public UserSubscriptionModel(int id, int subscribee_id, int subscriber_id)
            : base(id, subscriber_id)
        {
            this.Subscribee = new Lazy<ResearcherModel>(delegate() { return ResearcherModel.Find(subscribee_id); }, subscribee_id);
        }

        public static UserSubscriptionModel Create(int subscribee_id, int subscriber_id)
        {
            return DatabaseFacade.ExecuteTypedTx<UserSubscriptionModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT sub_id FROM ResearcherSubscriptions WHERE subscribee_id = @Subscribee AND subscriber_id = @Subscriber";
                cmd.Parameters.Add("@Subscribee", SqlDbType.Int).Value = subscribee_id;
                cmd.Parameters.Add("@Subscriber", SqlDbType.Int).Value = subscriber_id;
                cmd.Prepare();

                var id = cmd.ExecuteScalar();
                if (id == null)
                {
                    cmd.CommandText = "INSERT INTO ResearcherSubscriptions (subscribee_id, subscriber_id) OUTPUT INSERTED.sub_id VALUES (@Subscribee, @Subscriber)";
                    cmd.Prepare();

                    id = cmd.ExecuteScalar();
                }
                return new UserSubscriptionModel((Int32)id, subscribee_id, subscriber_id);
            });
        }

        public static void Delete(int id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ResearcherSubscriptions WHERE sub_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        /**
         * Checks if subscriber_id is subscribed to subscribee_id.
         */
        public static bool IsSubscribed(int subscribee_id, int subscriber_id)
        {
            return DatabaseFacade.ExecuteTypedTx<bool>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT sub_id FROM ResearcherSubscriptions WHERE subscribee_id = @Subscribee AND subscriber = @Subscriber";
                cmd.Parameters.Add("@Subscribee", SqlDbType.Int).Value = subscribee_id;
                cmd.Parameters.Add("@Subscriber", SqlDbType.Int).Value = subscriber_id; ;
                cmd.Prepare();
                return cmd.ExecuteScalar() != null;
            });
        }

        /**
         * Find all Users the input user_id is subscribed to.
         * E.g. If A -> B means that A is subscribed to B then this function finds all Bs for a given A.
         */
        public static List<UserSubscriptionModel> FindAllSubscriptions(int user_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<UserSubscriptionModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<UserSubscriptionModel>();

                cmd.CommandText = "SELECT sub_id, subscribee_id, subscriber_id FROM ResearcherSubscriptions WHERE subscriber_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = user_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new UserSubscriptionModel(rdr.GetInt32(rdr.GetOrdinal("sub_id")),
                                                             rdr.GetInt32(rdr.GetOrdinal("subscribee_id")),
                                                             rdr.GetInt32(rdr.GetOrdinal("subscriber_id"))));
                    }
                }

                return result;
            });
        }

        /**
         * Find all Users who are subscribed to the given user_id.
         * E.g if A -> B means that A is subscribed to B then this function finds all As for a given B.
         */
        public static List<UserSubscriptionModel> FindAllSubscribers(int user_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<UserSubscriptionModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<UserSubscriptionModel>();

                cmd.CommandText = "SELECT sub_id, subscribee_id, subscriber_id FROM ResearcherSubscriptions WHERE subscribee_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = user_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new UserSubscriptionModel(rdr.GetInt32(rdr.GetOrdinal("sub_id")),
                                                             rdr.GetInt32(rdr.GetOrdinal("subscribee_id")),
                                                             rdr.GetInt32(rdr.GetOrdinal("subscriber_id"))));
                    }
                }

                return result;
            });
        }

        public override SubscriptionKind GetKind() { return SubscriptionKind.User; }
    }

    public class ProgramSubscriptionModel : SubscriptionModel
    {
        public Lazy<ProgramModel> Program;

        public ProgramSubscriptionModel(int id, int program_id, int subscriber_id)
            : base(id, subscriber_id)
        {
            this.Program = new Lazy<ProgramModel>(delegate() { return ProgramModel.Find(program_id); }, program_id);
        }

        public static ProgramSubscriptionModel Create(int program_id, int subscriber_id)
        {
            return DatabaseFacade.ExecuteTypedTx<ProgramSubscriptionModel>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT sub_id FROM ProgramSubscriptions WHERE program_id = @Program AND subscriber_id = @Subscriber";
                cmd.Parameters.Add("@Program", SqlDbType.Int).Value = program_id;
                cmd.Parameters.Add("@Subscriber", SqlDbType.Int).Value = subscriber_id;
                cmd.Prepare();

                var id = cmd.ExecuteScalar();
                if (id == null)
                {
                    cmd.CommandText = "INSERT INTO ProgramSubscriptions (program_id, subscriber_id) OUTPUT INSERTED.sub_id VALUES (@Program, @Subscriber)";
                    cmd.Prepare();

                    id = cmd.ExecuteScalar();
                }
                return new ProgramSubscriptionModel((Int32)id, program_id, subscriber_id);
            });
        }

        public static void Delete(int id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ProgramSubscriptions WHERE sub_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        public static void Delete(int program_id, int subscriber_id)
        {
            DatabaseFacade.ExecuteTx(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "DELETE FROM ProgramSubscriptions WHERE program_id = @Program AND subscriber_id = @Subscriber";
                cmd.Parameters.Add("@Program", SqlDbType.Int).Value = program_id;
                cmd.Parameters.Add("@Subscriber", SqlDbType.Int).Value = subscriber_id;
                cmd.Prepare();
                return cmd.ExecuteNonQuery() != 0;
            });
        }

        /**
         * Checks if subscriber_id is subscribed to program_id.
         */
        public static bool IsSubscribed(int program_id, int subscriber_id)
        {
            return DatabaseFacade.ExecuteTypedTx<bool>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                cmd.CommandText = "SELECT sub_id FROM ProgramSubscriptions WHERE program_id = @Program AND subscriber_id = @Subscriber";
                cmd.Parameters.Add("@Program", SqlDbType.Int).Value = program_id;
                cmd.Parameters.Add("@Subscriber", SqlDbType.Int).Value = subscriber_id;
                cmd.Prepare();
                return cmd.ExecuteScalar() != null;
            });
        }

        /**
         * Find all Users the input user_id is subscribed to.
         * E.g. If A -> B means that A is subscribed to B then this function finds all Bs for a given A.
         */
        public static List<ProgramSubscriptionModel> FindAllSubscriptions(int user_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<ProgramSubscriptionModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<ProgramSubscriptionModel>();

                cmd.CommandText = "SELECT sub_id, program_id, subscriber_id FROM ProgramSubscriptions WHERE subscriber_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = user_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new ProgramSubscriptionModel(rdr.GetInt32(rdr.GetOrdinal("sub_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("program_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("subscriber_id"))));
                    }
                }

                return result;
            });
        }

        /**
         * Find all Users who are subscribed to the given user_id.
         * E.g if A -> B means that A is subscribed to B then this function finds all As for a given B.
         */
        public static List<ProgramSubscriptionModel> FindAllSubscribers(int program_id)
        {
            return DatabaseFacade.ExecuteTypedTx<List<ProgramSubscriptionModel>>(delegate(SqlConnection conn, SqlTransaction tx, SqlCommand cmd)
            {
                var result = new List<ProgramSubscriptionModel>();

                cmd.CommandText = "SELECT sub_id, program_id, subscriber_id FROM ProgramSubscriptions WHERE program_id = @Id";
                cmd.Parameters.Add("@Id", SqlDbType.Int).Value = program_id;
                cmd.Prepare();

                using (var rdr = cmd.ExecuteReader())
                {
                    while (rdr.Read())
                    {
                        result.Add(new ProgramSubscriptionModel(rdr.GetInt32(rdr.GetOrdinal("sub_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("program_id")),
                                                                rdr.GetInt32(rdr.GetOrdinal("subscriber_id"))));
                    }
                }

                return result;
            });
        }

        public override SubscriptionKind GetKind() { return SubscriptionKind.Program; }
    }
}